/* ////////////////////////////////////////////////////////////////////////////
 * File:        comstl_interface_traits.h
 *
 * Purpose:     Interface traits.
 *
 * Created:     25th May 2002
 * Updated:     30th May 2004
 *
 * Author:      Matthew Wilson, Synesis Software Pty Ltd.
 *
 * License:     (Licensed under the Synesis Software Standard Source License)
 *
 *              Copyright (C) 2002-2004, Synesis Software Pty Ltd.
 *
 *              All rights reserved.
 *
 *              www:        http://www.synesis.com.au/comstl
 *                          http://www.comstl.org/
 *
 *              email:      submissions@comstl.org  for submissions
 *                          admin@comstl.org        for other enquiries
 *
 *              Redistribution and use in source and binary forms, with or
 *              without modification, are permitted provided that the following
 *              conditions are met:
 *
 *              (i) Redistributions of source code must retain the above
 *              copyright notice and contact information, this list of
 *              conditions and the following disclaimer.
 *
 *              (ii) Any derived versions of this software (howsoever modified)
 *              remain the sole property of Synesis Software.
 *
 *              (iii) Any derived versions of this software (howsoever modified)
 *              remain subject to all these conditions.
 *
 *              (iv) Neither the name of Synesis Software nor the names of any
 *              subdivisions, employees or agents of Synesis Software, nor the
 *              names of any other contributors to this software may be used to
 *              endorse or promote products derived from this software without
 *              specific prior written permission.
 *
 *              This source code is provided by Synesis Software "as is" and any
 *              warranties, whether expressed or implied, including, but not
 *              limited to, the implied warranties of merchantability and
 *              fitness for a particular purpose are disclaimed. In no event
 *              shall the Synesis Software be liable for any direct, indirect,
 *              incidental, special, exemplary, or consequential damages
 *              (including, but not limited to, procurement of substitute goods
 *              or services; loss of use, data, or profits; or business
 *              interruption) however caused and on any theory of liability,
 *              whether in contract, strict liability, or tort (including
 *              negligence or otherwise) arising in any way out of the use of
 *              this software, even if advised of the possibility of such
 *              damage.
 *
 * ////////////////////////////////////////////////////////////////////////// */


/// \file comstl_interface_traits.h
///
/// Interface traits.

#ifndef COMSTL_INCL_H_COMSTL_INTERFACE_TRAITS
#define COMSTL_INCL_H_COMSTL_INTERFACE_TRAITS

#ifndef __STLSOFT_DOCUMENTATION_SKIP_SECTION
# define COMSTL_VER_H_COMSTL_INTERFACE_TRAITS_MAJOR     1
# define COMSTL_VER_H_COMSTL_INTERFACE_TRAITS_MINOR     3
# define COMSTL_VER_H_COMSTL_INTERFACE_TRAITS_REVISION  5
# define COMSTL_VER_H_COMSTL_INTERFACE_TRAITS_EDIT      33
#endif /* !__STLSOFT_DOCUMENTATION_SKIP_SECTION */

/* ////////////////////////////////////////////////////////////////////////////
 * Includes
 */

#ifndef COMSTL_INCL_H_COMSTL
# include "comstl.h"    // Include the COMSTL root header
#endif /* !COMSTL_INCL_H_COMSTL */

/* /////////////////////////////////////////////////////////////////////////////
 * Namespace
 *
 * The COMSTL components are contained within the comstl namespace. This is
 * actually an alias for stlsoft::comstl_project,
 *
 * The definition matrix is as follows:
 *
 * _STLSOFT_NO_NAMESPACE    _COMSTL_NO_NAMESPACE    comstl definition
 * ---------------------    --------------------    -----------------
 *  not defined              not defined             = stlsoft::comstl_project
 *  not defined              defined                 not defined
 *  defined                  not defined             comstl
 *  defined                  defined                 not defined
 *
 */

#ifndef _COMSTL_NO_NAMESPACE
# if defined(_STLSOFT_NO_NAMESPACE) || \
     defined(__STLSOFT_DOCUMENTATION_SKIP_SECTION)
/* There is no stlsoft namespace, so must define ::comstl */
namespace comstl
{
# else
/* Define stlsoft::comstl_project */

namespace stlsoft
{

namespace comstl_project
{

# endif /* _STLSOFT_NO_NAMESPACE */
#endif /* !_COMSTL_NO_NAMESPACE */

/* ////////////////////////////////////////////////////////////////////////// */

/// \defgroup comstl_itf_library Interface Library
/// \ingroup COMSTL libraries
/// \brief This library provides facilities for manipulating COM interfaces
/// @{

/* ////////////////////////////////////////////////////////////////////////////
 * Macros
 */

#ifndef __STLSOFT_DOCUMENTATION_SKIP_SECTION

#define COMSTL_IID_TRAITS_DEFINE__(I, T) \
  STLSOFT_TEMPLATE_SPECIALISATION \
struct IID_traits<T> \
{ \
public: \
    static REFIID   iid() { return IID_##I; } \
};

#define _COMSTL_IID_TRAITS_DEFINE(I)        COMSTL_IID_TRAITS_DEFINE__(I, I)

#endif /* !__STLSOFT_DOCUMENTATION_SKIP_SECTION */

/* ////////////////////////////////////////////////////////////////////////////
 * Classes
 */

#ifdef __STLSOFT_DOCUMENTATION_SKIP_SECTION

/// Interface identifier traits
///
/// This traits class takes an interface type and provides a specialisation that
/// has a statid iid() method which returns the interface identifier (IID) for
/// that type.
///
/// To use, simply have the expression IID_traits<I>::iid() where I is your
/// interface type.
///
/// On compilers that support the __uuidof pseudo-operator this is used,
/// otherwise you must specify specialisations - using the
/// COMSTL_IID_TRAITS_DEFINE macro - for your interfaces, as in
///
///   COMSTL_IID_TRAITS_DEFINE(IMyInterface)
///
/// All the interfaces currently defined in unknown.idl and objidl.idl are
/// so defined in the file comstl_interface_traits_std.h, which is included in
/// for non-__uuidof compilations.
///
/// \param I The interface
template <class I>
struct IID_traits
{
public:
    /// Returns a reference to the IID for the parameterising interface
    static REFIID   iid();
};
#else

#if !defined(_COMSTL_NO_UUIDOF) && \
    (   defined(__STLSOFT_COMPILER_IS_BORLAND) ||  \
        defined(__STLSOFT_COMPILER_IS_INTEL) ||  \
        defined(__STLSOFT_COMPILER_IS_MWERKS) ||  \
        (   defined(__STLSOFT_COMPILER_IS_MSVC) && \
            _MSC_VER >= 1200))
template <class I>
struct IID_traits
{
public:
    static REFIID   iid() { return __uuidof(I); }
};

#define COMSTL_IID_TRAITS_DEFINE(I)

#else
template <class I>
struct IID_traits;

#define COMSTL_IID_TRAITS_DEFINE(I)     EXTERN_C const IID IID_##I; \
                                        _COMSTL_IID_TRAITS_DEFINE(I) \
                                        COMSTL_IID_TRAITS_DEFINE__(I, I*)

/* For backwards compatibility */
#define comstl_IID_traits_define(I)     COMSTL_IID_TRAITS_DEFINE(I)

#endif /* __STLSOFT_DOCUMENTATION_SKIP_SECTION */

#ifndef _COMSTL_NO_NAMESPACE
# if defined(_STLSOFT_NO_NAMESPACE) || \
     defined(__STLSOFT_DOCUMENTATION_SKIP_SECTION)
} // namespace comstl
# else
} // namespace comstl_project
} // namespace stlsoft
# endif /* _STLSOFT_NO_NAMESPACE */
#endif /* !_COMSTL_NO_NAMESPACE */

#include "comstl_interface_traits_std.h"

#ifndef _COMSTL_NO_NAMESPACE
# if defined(_STLSOFT_NO_NAMESPACE) || \
     defined(__STLSOFT_DOCUMENTATION_SKIP_SECTION)
/* There is no stlsoft namespace, so must define ::comstl */
namespace comstl
{
# else
/* Define stlsoft::comstl_project */

namespace stlsoft
{

namespace comstl_project
{

# endif /* _STLSOFT_NO_NAMESPACE */
#endif /* !_COMSTL_NO_NAMESPACE */

#endif /* __SYNSOFT_DVS_COMPILER_VER_MSVC <
__SYNSOFT_GEN_COMPILER_VERSION(5, 0) */

/* ////////////////////////////////////////////////////////////////////////// */

/// @} // end of group comstl_itf_library

/* ////////////////////////////////////////////////////////////////////////// */

#ifndef _COMSTL_NO_NAMESPACE
# if defined(_STLSOFT_NO_NAMESPACE) || \
     defined(__STLSOFT_DOCUMENTATION_SKIP_SECTION)
} // namespace comstl
# else
} // namespace comstl_project
} // namespace stlsoft
# endif /* _STLSOFT_NO_NAMESPACE */
#endif /* !_COMSTL_NO_NAMESPACE */

/* ////////////////////////////////////////////////////////////////////////// */

#endif /* !COMSTL_INCL_H_COMSTL_INTERFACE_TRAITS */

/* ////////////////////////////////////////////////////////////////////////// */
