/* /////////////////////////////////////////////////////////////////////////////
 * File:        comstl_security_initialisers.h (originally MOSecFns.h, ::SynesisCom)
 *
 * Purpose:     Contains classes for initialising COM/OLE.
 *
 * Created:     1st February 1998
 * Updated:     30th May 2004
 *
 * Author:      Matthew Wilson, Synesis Software Pty Ltd.
 *
 * License:     (Licensed under the Synesis Software Standard Source License)
 *
 *              Copyright (C) 2002-2004, Synesis Software Pty Ltd.
 *
 *              All rights reserved.
 *
 *              www:        http://www.synesis.com.au/comstl
 *                          http://www.comstl.org/
 *
 *              email:      submissions@comstl.org  for submissions
 *                          admin@comstl.org        for other enquiries
 *
 *              Redistribution and use in source and binary forms, with or
 *              without modification, are permitted provided that the following
 *              conditions are met:
 *
 *              (i) Redistributions of source code must retain the above
 *              copyright notice and contact information, this list of
 *              conditions and the following disclaimer.
 *
 *              (ii) Any derived versions of this software (howsoever modified)
 *              remain the sole property of Synesis Software.
 *
 *              (iii) Any derived versions of this software (howsoever modified)
 *              remain subject to all these conditions.
 *
 *              (iv) Neither the name of Synesis Software nor the names of any
 *              subdivisions, employees or agents of Synesis Software, nor the
 *              names of any other contributors to this software may be used to
 *              endorse or promote products derived from this software without
 *              specific prior written permission.
 *
 *              This source code is provided by Synesis Software "as is" and any
 *              warranties, whether expressed or implied, including, but not
 *              limited to, the implied warranties of merchantability and
 *              fitness for a particular purpose are disclaimed. In no event
 *              shall the Synesis Software be liable for any direct, indirect,
 *              incidental, special, exemplary, or consequential damages
 *              (including, but not limited to, procurement of substitute goods
 *              or services; loss of use, data, or profits; or business
 *              interruption) however caused and on any theory of liability,
 *              whether in contract, strict liability, or tort (including
 *              negligence or otherwise) arising in any way out of the use of
 *              this software, even if advised of the possibility of such
 *              damage.
 *
 * ////////////////////////////////////////////////////////////////////////// */


/// \file comstl_security_initialisers.h
///
/// Contains classes for initialising COM/OLE.

#ifndef COMSTL_INCL_H_COMSTL_SECURITY_INITIALISERS
#define COMSTL_INCL_H_COMSTL_SECURITY_INITIALISERS

#ifndef __STLSOFT_DOCUMENTATION_SKIP_SECTION
# define COMSTL_VER_H_COMSTL_SECURITY_INITIALISERS_MAJOR        2
# define COMSTL_VER_H_COMSTL_SECURITY_INITIALISERS_MINOR        2
# define COMSTL_VER_H_COMSTL_SECURITY_INITIALISERS_REVISION     3
# define COMSTL_VER_H_COMSTL_SECURITY_INITIALISERS_EDIT         19
#endif /* !__STLSOFT_DOCUMENTATION_SKIP_SECTION */

/* ////////////////////////////////////////////////////////////////////////////
 * Includes
 */

#ifndef COMSTL_INCL_H_COMSTL
# include "comstl.h"                // Include the COMSTL root header
#endif /* !COMSTL_INCL_H_COMSTL */

#if (   !defined(_WIN32_WINNT) || \
        (_WIN32_WINNT < 0x0400)) && \
    !defined(_WIN32_DCOM)
# error This file can only be used in the context of DCOM compilations.
#endif /* _WIN32_WINNT < 0x0400 && !_WIN32_DCOM */

#if _MSC_VER >= 1200
# include <iaccess.h>
#endif /* _MSC_VER >= 1200 */

/* /////////////////////////////////////////////////////////////////////////////
 * Namespace
 *
 * The COMSTL components are contained within the comstl namespace. This is
 * actually an alias for stlsoft::comstl_project,
 *
 * The definition matrix is as follows:
 *
 * _STLSOFT_NO_NAMESPACE    _COMSTL_NO_NAMESPACE    comstl definition
 * ---------------------    --------------------    -----------------
 *  not defined              not defined             = stlsoft::comstl_project
 *  not defined              defined                 not defined
 *  defined                  not defined             comstl
 *  defined                  defined                 not defined
 *
 */

#ifndef _COMSTL_NO_NAMESPACE
# if defined(_STLSOFT_NO_NAMESPACE) || \
     defined(__STLSOFT_DOCUMENTATION_SKIP_SECTION)
/* There is no stlsoft namespace, so must define ::comstl */
namespace comstl
{
# else
/* Define stlsoft::comstl_project */

namespace stlsoft
{

namespace comstl_project
{

# endif /* _STLSOFT_NO_NAMESPACE */
#endif /* !_COMSTL_NO_NAMESPACE */

/* ////////////////////////////////////////////////////////////////////////// */

/// \defgroup comstl_security_intialisers Security Initialiser Functions
/// \ingroup COMSTL
/// \brief These functions initialise the COM security libraries
/// @{

/* /////////////////////////////////////////////////////////////////////////////
 * Constants & definitions.
 */

#ifndef __STLSOFT_DOCUMENTATION_SKIP_SECTION
const DWORD _EOAC_SEC_DESC          =   0x0;
const DWORD _EOAC_ACCESS_CONTROL    =   0x4;
const DWORD _EOAC_APPID             =   0x8;
const DWORD _EOAC_SECURITY_MASK     =   (   _EOAC_SEC_DESC |
                                            _EOAC_APPID |
                                            _EOAC_ACCESS_CONTROL);
#endif /* !__STLSOFT_DOCUMENTATION_SKIP_SECTION */

/* /////////////////////////////////////////////////////////////////////////////
 * Functions
 */

// function CoInitSecurity
//
// The three overloaded CoInitSecurity functions
// provide a type-safe interface to the CoInitializeSecurity
// API, discriminating between IAccessControl, AppID, and
// SECURITY_DESCRIPTOR security information parameters.

/// Initialises the COM security libraries with an IAccessControl instance
///
#ifdef __IAccessControl_INTERFACE_DEFINED__
inline HRESULT CoInitSecurity(  LPUNKNOWN                   punkAccessControl,
                                DWORD                       cAuthSvc,
                                SOLE_AUTHENTICATION_SERVICE *asAuthSvc,
                                DWORD                       dwAuthnLevel,
                                DWORD                       dwImpLevel,
                                RPC_AUTH_IDENTITY_HANDLE    pAuthInfo,
                                DWORD                       dwCapabilities)
{
    HRESULT         hr;
    IAccessControl  *pac;

    /* Ensure correct flag. */
    dwCapabilities &= ~_EOAC_SECURITY_MASK;
    dwCapabilities |= _EOAC_ACCESS_CONTROL;

    hr = punkAccessControl->QueryInterface(IID_IAccessControl, (void**)&pac);

    if(SUCCEEDED(hr))
    {
        hr = ::CoInitializeSecurity(pac, cAuthSvc, asAuthSvc, NULL, dwAuthnLevel, dwImpLevel, pAuthInfo, dwCapabilities, NULL);

        pac->Release();
    }

    return hr;
}
#endif // __IAccessControl_INTERFACE_DEFINED__

/// Initialises the COM security libraries with an APPID
///
inline HRESULT CoInitSecurity(  const GUID                  &appid,
                                DWORD                       cAuthSvc,
                                SOLE_AUTHENTICATION_SERVICE *asAuthSvc,
                                DWORD                       dwAuthnLevel,
                                DWORD                       dwImpLevel,
                                RPC_AUTH_IDENTITY_HANDLE    pAuthInfo,
                                DWORD                       dwCapabilities)
{
    /* Ensure correct flag. */
    dwCapabilities &= ~_EOAC_SECURITY_MASK;
    dwCapabilities |= _EOAC_APPID;

    return ::CoInitializeSecurity((LPVOID)&appid, cAuthSvc, asAuthSvc, NULL, dwAuthnLevel, dwImpLevel, pAuthInfo, dwCapabilities, NULL);
}

/// Initialises the COM security libraries with a security descriptor
///
inline HRESULT CoInitSecurity(  SECURITY_DESCRIPTOR         *psd,
                                DWORD                       cAuthSvc,
                                SOLE_AUTHENTICATION_SERVICE *asAuthSvc,
                                DWORD                       dwAuthnLevel,
                                DWORD                       dwImpLevel,
                                RPC_AUTH_IDENTITY_HANDLE    pAuthInfo,
                                DWORD                       dwCapabilities)
{
    /* Ensure correct flag. */
    dwCapabilities &= ~_EOAC_SECURITY_MASK;
    dwCapabilities |= _EOAC_SEC_DESC;

    return ::CoInitializeSecurity(psd, cAuthSvc, asAuthSvc, NULL, dwAuthnLevel, dwImpLevel, pAuthInfo, dwCapabilities, NULL);
}

////////////////////////////////////////////////////////////////////////////////
// Unit-testing

#ifdef STLSOFT_UNITTEST

namespace unittest
{
    ss_bool_t test_comstl_security_initialisers(unittest_reporter *r)
    {
        using stlsoft::unittest::unittest_initialiser;

        ss_bool_t               bSuccess    =   true;

        unittest_initialiser    init(r, "COMSTL", "security_initialisers", __FILE__);

#if 0
        if(<<TODO>>)
        {
            r->report("<<TODO>> failed ", __LINE__);
            bSuccess = false;
        }
#endif /* 0 */

        return bSuccess;
    }

    unittest_registrar    unittest_comstl_security_initialisers(test_comstl_security_initialisers);

} // namespace unittest

#endif /* STLSOFT_UNITTEST */

/* ////////////////////////////////////////////////////////////////////////// */

/// @} // end of group comstl_security_intialisers

/* ////////////////////////////////////////////////////////////////////////// */

#ifndef _COMSTL_NO_NAMESPACE
# if defined(_STLSOFT_NO_NAMESPACE) || \
     defined(__STLSOFT_DOCUMENTATION_SKIP_SECTION)
} // namespace comstl
# else
} // namespace stlsoft::comstl_project
} // namespace stlsoft
# endif /* _STLSOFT_NO_NAMESPACE */
#endif /* !_COMSTL_NO_NAMESPACE */

/* ////////////////////////////////////////////////////////////////////////// */

#endif /* !COMSTL_INCL_H_COMSTL_SECURITY_INITIALISERS */

/* ////////////////////////////////////////////////////////////////////////// */
