/* /////////////////////////////////////////////////////////////////////////////
 * File:        dotnetstl_object_enumerators.h
 *
 * Purpose:     A simple object enumerator, based on ArrayList.
 *
 * Created:     24th June 2003
 * Updated:     10th June 2004
 *
 * Author:      Matthew Wilson, Synesis Software Pty Ltd.
 *
 * License:     (Licensed under the Synesis Software Standard Source License)
 *
 *              Copyright (C) 2002-2004, Synesis Software Pty Ltd.
 *
 *              All rights reserved.
 *
 *              www:        http://www.synesis.com.au/dotnetstl
 *                          http://www.dotnetstl.org/
 *
 *              email:      submissions@dotnetstl.org  for submissions
 *                          admin@dotnetstl.org        for other enquiries
 *
 *              Redistribution and use in source and binary forms, with or
 *              without modification, are permitted provided that the following
 *              conditions are met:
 *
 *              (i) Redistributions of source code must retain the above
 *              copyright notice and contact information, this list of
 *              conditions and the following disclaimer.
 *
 *              (ii) Any derived versions of this software (howsoever modified)
 *              remain the sole property of Synesis Software.
 *
 *              (iii) Any derived versions of this software (howsoever modified)
 *              remain subject to all these conditions.
 *
 *              (iv) Neither the name of Synesis Software nor the names of any
 *              subdivisions, employees or agents of Synesis Software, nor the
 *              names of any other contributors to this software may be used to
 *              endorse or promote products derived from this software without
 *              specific prior written permission.
 *
 *              This source code is provided by Synesis Software "as is" and any
 *              warranties, whether expressed or implied, including, but not
 *              limited to, the implied warranties of merchantability and
 *              fitness for a particular purpose are disclaimed. In no event
 *              shall the Synesis Software be liable for any direct, indirect,
 *              incidental, special, exemplary, or consequential damages
 *              (including, but not limited to, procurement of substitute goods
 *              or services; loss of use, data, or profits; or business
 *              interruption) however caused and on any theory of liability,
 *              whether in contract, strict liability, or tort (including
 *              negligence or otherwise) arising in any way out of the use of
 *              this software, even if advised of the possibility of such
 *              damage.
 *
 * ////////////////////////////////////////////////////////////////////////// */


/// \file dotnetstl_object_enumerators.h
///
/// A simple object enumerator, based on ArrayList.

#ifndef DOTNETSTL_INCL_H_DOTNETSTL_OBJECT_ENUMERATORS
#define DOTNETSTL_INCL_H_DOTNETSTL_OBJECT_ENUMERATORS

#ifndef __STLSOFT_DOCUMENTATION_SKIP_SECTION
# define DOTNETSTL_VER_H_DOTNETSTL_OBJECT_ENUMERATORS_MAJOR     1
# define DOTNETSTL_VER_H_DOTNETSTL_OBJECT_ENUMERATORS_MINOR     0
# define DOTNETSTL_VER_H_DOTNETSTL_OBJECT_ENUMERATORS_REVISION  6
# define DOTNETSTL_VER_H_DOTNETSTL_OBJECT_ENUMERATORS_EDIT      10
#endif /* !__STLSOFT_DOCUMENTATION_SKIP_SECTION */

/* /////////////////////////////////////////////////////////////////////////////
 * Includes
 */

#ifndef DOTNETSTL_INCL_H_DOTNETSTL
# include "dotnetstl.h"                     // Include the .netSTL root header
#endif /* !DOTNETSTL_INCL_H_DOTNETSTL */

/* ////////////////////////////////////////////////////////////////////////// */

#ifdef _STLSOFT_NO_NAMESPACE
/* There is no stlsoft namespace, so must define ::dotnetstl */
namespace dotnetstl
{
#else
/* Define stlsoft::dotnet_project */

namespace stlsoft
{

namespace dotnetstl_project
{

#endif /* _STLSOFT_NO_NAMESPACE */

/* /////////////////////////////////////////////////////////////////////////////
 * Classes
 */

/// Provides an IEnumerator implementation on an ArrayList
public __gc class ArrayListEnumerator
    : public System::Collections::IEnumerator
{
//    typedef System::Collections::ArrayList  ArrayList;

public:
    /// Creates an enumerator for the given list.
    ArrayListEnumerator(System::Collections::ArrayList *objects)
        : m_objects(objects)
        , m_index(-1)
    {}

public: // VC++ 7.1 requires that these be public
    /// Returns the current object
    Object* get_Current()
    {
        if( m_index < 0 ||
            !(m_index < m_objects->get_Count()))
        {
            throw new System::InvalidOperationException();
        }

        return m_objects->get_Item(m_index);
    }

    /// Advances the search by one position
    ///
    /// \retval true An object is available at the new position
    /// \retval false The enumeration is complete
    bool MoveNext()
    {
        if(m_index < m_objects->get_Count() - 1)
        {
            ++m_index;

            return true;
        }
        else
        {
            return false;
        }
    }

    /// Resets the enumeration
    void Reset()
    {
        m_index = -1;
    }

// Members
private:
    System::Collections::ArrayList  *m_objects;
    int                             m_index;

// Not to be implemented
private:
    ArrayListEnumerator(ArrayListEnumerator const &)
    {}
//  ArrayListEnumerator &operator =(ArrayListEnumerator const &);
};

/* ////////////////////////////////////////////////////////////////////////// */

#ifdef _STLSOFT_NO_NAMESPACE
} // namespace dotnetstl
#else
} // namespace dotnetstl_project
} // namespace stlsoft
#endif /* _STLSOFT_NO_NAMESPACE */

/* ////////////////////////////////////////////////////////////////////////// */

#endif /* DOTNETSTL_INCL_H_DOTNETSTL_OBJECT_ENUMERATORS */

/* ////////////////////////////////////////////////////////////////////////// */
