/* /////////////////////////////////////////////////////////////////////////////
 * File:        stlsoft_string_access_exception.h
 *
 * Purpose:     Contains the string access shims for std::exception.
 *
 * Created:     2nd May 2003
 * Updated:     30th May 2004
 *
 * Author:      Matthew Wilson, Synesis Software Pty Ltd.
 *
 * License:     (Licensed under the Synesis Software Standard Source License)
 *
 *              Copyright (C) 2002-2003, Synesis Software Pty Ltd.
 *
 *              All rights reserved.
 *
 *              www:        http://www.synesis.com.au/stlsoft
 *                          http://www.stlsoft.org/
 *
 *              email:      submissions@stlsoft.org  for submissions
 *                          admin@stlsoft.org        for other enquiries
 *
 *              Redistribution and use in source and binary forms, with or
 *              without modification, are permitted provided that the following
 *              conditions are met:
 *
 *              (i) Redistributions of source code must retain the above
 *              copyright notice and contact information, this list of
 *              conditions and the following disclaimer.
 *
 *              (ii) Any derived versions of this software (howsoever modified)
 *              remain the sole property of Synesis Software.
 *
 *              (iii) Any derived versions of this software (howsoever modified)
 *              remain subject to all these conditions.
 *
 *              (iv) Neither the name of Synesis Software nor the names of any
 *              subdivisions, employees or agents of Synesis Software, nor the
 *              names of any other contributors to this software may be used to
 *              endorse or promote products derived from this software without
 *              specific prior written permission.
 *
 *              This source code is provided by Synesis Software "as is" and any
 *              warranties, whether expressed or implied, including, but not
 *              limited to, the implied warranties of merchantability and
 *              fitness for a particular purpose are disclaimed. In no event
 *              shall the Synesis Software be liable for any direct, indirect,
 *              incidental, special, exemplary, or consequential damages
 *              (including, but not limited to, procurement of substitute goods
 *              or services; loss of use, data, or profits; or business
 *              interruption) however caused and on any theory of liability,
 *              whether in contract, strict liability, or tort (including
 *              negligence or otherwise) arising in any way out of the use of
 *              this software, even if advised of the possibility of such
 *              damage.
 *
 * ////////////////////////////////////////////////////////////////////////// */


/// \file stlsoft_string_access_exception.h
///
/// Contains the string access shims for std::exception.

#ifndef STLSOFT_INCL_H_STLSOFT_STRING_ACCESS_EXCEPTION
#define STLSOFT_INCL_H_STLSOFT_STRING_ACCESS_EXCEPTION

#ifndef __STLSOFT_DOCUMENTATION_SKIP_SECTION
#define _STLSOFT_VER_H_STLSOFT_STRING_ACCESS_EXCEPTION_MAJOR    1
#define _STLSOFT_VER_H_STLSOFT_STRING_ACCESS_EXCEPTION_MINOR    0
#define _STLSOFT_VER_H_STLSOFT_STRING_ACCESS_EXCEPTION_REVISION 2
#define _STLSOFT_VER_H_STLSOFT_STRING_ACCESS_EXCEPTION_EDIT     7
#endif /* !__STLSOFT_DOCUMENTATION_SKIP_SECTION */

/* /////////////////////////////////////////////////////////////////////////////
 * Includes
 */

#ifndef STLSOFT_INCL_H_STLSOFT
# include "stlsoft.h"               // Include the STLSoft root header
#endif /* !STLSOFT_INCL_H_STLSOFT */
#ifndef STLSOFT_INCL_H_STLSOFT_STRING_ACCESS
# include "stlsoft_string_access.h" // Include the standard string access header
#endif /* !STLSOFT_INCL_H_STLSOFT_STRING_ACCESS */
#include <exception>                // std::exception

/* /////////////////////////////////////////////////////////////////////////////
 * Namespace
 */

#ifndef _STLSOFT_NO_NAMESPACE
namespace stlsoft
{
#endif /* _STLSOFT_NO_NAMESPACE */

/* ////////////////////////////////////////////////////////////////////////// */

/// \weakgroup concepts STLSoft Concepts

/// \weakgroup concepts_shims Shims
/// \ingroup concepts

/// \weakgroup concepts_shims_string_access String Access Shims
/// \ingroup concepts_shims
/// \brief These \ref concepts_shims "shims" retrieve the C-string for arbitrary types

/// \defgroup stlsoft_string_access_shims String Access Shims (STLSoft)
/// \ingroup STLSoft concepts_shims_string_access
/// \brief These \ref concepts_shims "shims" retrieve the C-string for arbitrary types
/// @{

/* /////////////////////////////////////////////////////////////////////////////
 * c_str_ptr_null
 *
 * This can be applied to an expression, and the return value is either a
 * pointer to the character string or NULL.
 */

/* std::exception */

/// \brief Returns the corresponding C-string pointer of \c x, or a null pointer
inline ss_char_a_t const *c_str_ptr_null_a(stlsoft_ns_qual_std(exception) const &x)
{
    return c_str_ptr_null_a(x.what());
}

/// \brief Returns the corresponding C-string pointer of \c x, or a null pointer
inline ss_char_a_t const *c_str_ptr_null(stlsoft_ns_qual_std(exception) const &x)
{
    return c_str_ptr_null_a(x);
}

/* /////////////////////////////////////////////////////////////////////////////
 * c_str_ptr
 *
 * This can be applied to an expression, and the return value is either a
 * pointer to the character string or to an empty string.
 */

/* std::exception */
/// \brief Returns the corresponding C-string pointer of \c x
inline ss_char_a_t const *c_str_ptr_a(stlsoft_ns_qual_std(exception) const &x)
{
    return c_str_ptr_a(x.what());
}

/// \brief Returns the corresponding C-string pointer of \c x
inline ss_char_a_t const *c_str_ptr(stlsoft_ns_qual_std(exception) const &x)
{
    return c_str_ptr_a(x);
}

/* /////////////////////////////////////////////////////////////////////////////
 * c_str_len
 *
 * This can be applied to an expression, and the return value is the number of
 * characters in the character string in the expression.
 */

/* std::exception */
/// \brief Returns the length (in characters) of \c x, <b><i>not</i></b> including the null-terminating character
inline ss_size_t c_str_len_a(stlsoft_ns_qual_std(string) const &x)
{
    return c_str_len_a(x.what());
}

/// \brief Returns the length (in characters) of \c x, <b><i>not</i></b> including the null-terminating character
inline ss_size_t c_str_len(stlsoft_ns_qual_std(exception) const &x)
{
    return c_str_len_a(x);
}

/* /////////////////////////////////////////////////////////////////////////////
 * c_str_size
 *
 * This can be applied to an expression, and the return value is the number of
 * bytes required to store the character string in the expression, NOT including
 * the null-terminating character.
 */

/* std::exception */
/// \brief Returns the size (in bytes) of the contents of \c x, <b><i>not</i></b> including the null-terminating character
inline ss_size_t c_str_size_a(stlsoft_ns_qual_std(string) const &x)
{
    return c_str_size_a(x.what());
}

/// \brief Returns the size (in bytes) of the contents of \c x, <b><i>not</i></b> including the null-terminating character
inline ss_size_t c_str_size(stlsoft_ns_qual_std(exception) const &x)
{
    return c_str_size_a(x);
}

/* ////////////////////////////////////////////////////////////////////////// */

/// @} // end of group stlsoft_string_access_shims

/* ////////////////////////////////////////////////////////////////////////// */

#ifndef _STLSOFT_NO_NAMESPACE
} // namespace stlsoft
#endif /* _STLSOFT_NO_NAMESPACE */

/* ////////////////////////////////////////////////////////////////////////// */

#endif /* !STLSOFT_INCL_H_STLSOFT_STRING_ACCESS_EXCEPTION */

/* ////////////////////////////////////////////////////////////////////////// */
