/* /////////////////////////////////////////////////////////////////////////////
 * File:        stlsoft_any_caster.h
 *
 * Purpose:     A dangerous, but sometimes necessary, tool for handling bad
 *              libraries.
 *
 * Created:     12th May 2004
 * Updated:     10th June 2004
 *
 * Author:      Matthew Wilson, Synesis Software Pty Ltd.
 *
 * License:     (Licensed under the Synesis Software Standard Source License)
 *
 *              Copyright (C) 2002-2004, Synesis Software Pty Ltd.
 *
 *              All rights reserved.
 *
 *              www:        http://www.synesis.com.au/stlsoft
 *                          http://www.stlsoft.org/
 *
 *              email:      submissions@stlsoft.org  for submissions
 *                          admin@stlsoft.org        for other enquiries
 *
 *              Redistribution and use in source and binary forms, with or
 *              without modification, are permitted provided that the following
 *              conditions are met:
 *
 *              (i) Redistributions of source code must retain the above
 *              copyright notice and contact information, this list of
 *              conditions and the following disclaimer.
 *
 *              (ii) Any derived versions of this software (howsoever modified)
 *              remain the sole property of Synesis Software.
 *
 *              (iii) Any derived versions of this software (howsoever modified)
 *              remain subject to all these conditions.
 *
 *              (iv) Neither the name of Synesis Software nor the names of any
 *              subdivisions, employees or agents of Synesis Software, nor the
 *              names of any other contributors to this software may be used to
 *              endorse or promote products derived from this software without
 *              specific prior written permission.
 *
 *              This source code is provided by Synesis Software "as is" and any
 *              warranties, whether expressed or implied, including, but not
 *              limited to, the implied warranties of merchantability and
 *              fitness for a particular purpose are disclaimed. In no event
 *              shall the Synesis Software be liable for any direct, indirect,
 *              incidental, special, exemplary, or consequential damages
 *              (including, but not limited to, procurement of substitute goods
 *              or services; loss of use, data, or profits; or business
 *              interruption) however caused and on any theory of liability,
 *              whether in contract, strict liability, or tort (including
 *              negligence or otherwise) arising in any way out of the use of
 *              this software, even if advised of the possibility of such
 *              damage.
 *
 * ////////////////////////////////////////////////////////////////////////// */


/// \file stlsoft_any_caster.h
///
/// A dangerous, but sometimes necessary, tool for handling bad libraries.

#ifndef STLSOFT_INCL_H_STLSOFT_ANY_CASTER
#define STLSOFT_INCL_H_STLSOFT_ANY_CASTER

#ifndef __STLSOFT_DOCUMENTATION_SKIP_SECTION
# define STLSOFT_VER_H_STLSOFT_ANY_CASTER_MAJOR     1
# define STLSOFT_VER_H_STLSOFT_ANY_CASTER_MINOR     1
# define STLSOFT_VER_H_STLSOFT_ANY_CASTER_REVISION  2
# define STLSOFT_VER_H_STLSOFT_ANY_CASTER_EDIT      12
#endif /* !__STLSOFT_DOCUMENTATION_SKIP_SECTION */

/* /////////////////////////////////////////////////////////////////////////////
 * Includes
 */

#ifndef STLSOFT_INCL_H_STLSOFT
# include "stlsoft.h"   // Include the STLSoft root header
#endif /* !STLSOFT_INCL_H_STLSOFT */

/* /////////////////////////////////////////////////////////////////////////////
 * Namespace
 */

#ifndef _STLSOFT_NO_NAMESPACE
namespace stlsoft
{
#endif /* _STLSOFT_NO_NAMESPACE */

/* /////////////////////////////////////////////////////////////////////////////
 * Classes
 */

/// \brief a
template <ss_size_t N>
struct any_caster_invalid_type
{
    typedef any_caster_invalid_type<N>  class_type;

#if defined(__STLSOFT_COMPILER_IS_GCC)
protected:
#else /* ? compiler */
private:
#endif /* compiler */
    any_caster_invalid_type();
    any_caster_invalid_type(class_type const &);
};

/// \brief This class helps overcome problems when external libraries have errors in their
/// function parameter declarations, resulting in compilation-time type errors with some
/// versions of the libraries, and not with others.
///
/// \note See \link inetstl_filesystem_traits.h \endlink for an example of its use.
template<   typename    T
        ,   typename    T1
        ,   typename    T2
        ,   typename    T3  =   any_caster_invalid_type<3>*
        ,   typename    T4  =   any_caster_invalid_type<4>*
        ,   typename    T5  =   any_caster_invalid_type<5>*
        ,   typename    T6  =   any_caster_invalid_type<6>*
        ,   typename    T7  =   any_caster_invalid_type<7>*
        ,   typename    T8  =   any_caster_invalid_type<8>*
        >
union any_caster
{
    typedef any_caster<T, T1, T2, T3, T4, T5, T6, T7, T8>   class_type;

/// \name Construction
/// @{
public:
    /// Constructs an instance from the source type
    ss_explicit_k any_caster(T t)
    {
        this->t = t;
    }
/// @}

/// \name Conversion
/// @{
public:
    operator T1 () const
    {
        return t1;
    }
    operator T2 () const
    {
        return t2;
    }
    operator T3 () const
    {
        return t3;
    }
    operator T4 () const
    {
        return t4;
    }
    operator T5 () const
    {
        return t5;
    }
    operator T6 () const
    {
        return t6;
    }
    operator T7 () const
    {
        return t7;
    }
    operator T8 () const
    {
        return t8;
    }
/// @}

private:
    T   t;
    T1  t1;
    T2  t2;
    T3  t3;
    T4  t4;
    T5  t5;
    T6  t6;
    T7  t7;
    T8  t8;

// Not to be implemented
private:
    any_caster(class_type const &rhs);
    class_type &operator =(class_type const &rhs);
};

/* ////////////////////////////////////////////////////////////////////////// */

#ifndef _STLSOFT_NO_NAMESPACE
} // namespace stlsoft
#endif /* _STLSOFT_NO_NAMESPACE */

/* ////////////////////////////////////////////////////////////////////////// */

#endif /* !STLSOFT_INCL_H_STLSOFT_ANY_CASTER */

/* ////////////////////////////////////////////////////////////////////////// */
