/* /////////////////////////////////////////////////////////////////////////////
 * File:        stlsoft_explicit_cast_specialisations.h
 *
 * Purpose:     Specialisations of explicit_cast
 *
 * Created:     13th August 2003
 * Updated:     30th May 2004
 *
 * Author:      Matthew Wilson, Synesis Software Pty Ltd.
 *
 * License:     (Licensed under the Synesis Software Standard Source License)
 *
 *              Copyright (C) 2002-2004, Synesis Software Pty Ltd.
 *
 *              All rights reserved.
 *
 *              www:        http://www.synesis.com.au/stlsoft
 *                          http://www.stlsoft.org/
 *
 *              email:      submissions@stlsoft.org  for submissions
 *                          admin@stlsoft.org        for other enquiries
 *
 *              Redistribution and use in source and binary forms, with or
 *              without modification, are permitted provided that the following
 *              conditions are met:
 *
 *              (i) Redistributions of source code must retain the above
 *              copyright notice and contact information, this list of
 *              conditions and the following disclaimer.
 *
 *              (ii) Any derived versions of this software (howsoever modified)
 *              remain the sole property of Synesis Software.
 *
 *              (iii) Any derived versions of this software (howsoever modified)
 *              remain subject to all these conditions.
 *
 *              (iv) Neither the name of Synesis Software nor the names of any
 *              subdivisions, employees or agents of Synesis Software, nor the
 *              names of any other contributors to this software may be used to
 *              endorse or promote products derived from this software without
 *              specific prior written permission.
 *
 *              This source code is provided by Synesis Software "as is" and any
 *              warranties, whether expressed or implied, including, but not
 *              limited to, the implied warranties of merchantability and
 *              fitness for a particular purpose are disclaimed. In no event
 *              shall the Synesis Software be liable for any direct, indirect,
 *              incidental, special, exemplary, or consequential damages
 *              (including, but not limited to, procurement of substitute goods
 *              or services; loss of use, data, or profits; or business
 *              interruption) however caused and on any theory of liability,
 *              whether in contract, strict liability, or tort (including
 *              negligence or otherwise) arising in any way out of the use of
 *              this software, even if advised of the possibility of such
 *              damage.
 *
 * ////////////////////////////////////////////////////////////////////////// */


#ifndef STLSOFT_INCL_H_STLSOFT_EXPLICIT_CAST
# error This file is included from within stlsoft_explicit_cast.h, and cannot be included directly
#else

#ifndef __STLSOFT_DOCUMENTATION_SKIP_SECTION
# define STLSOFT_VER_H_EXPLICIT_CAST_SPECIALISATIONS_MAJOR      1
# define STLSOFT_VER_H_EXPLICIT_CAST_SPECIALISATIONS_MINOR      0
# define STLSOFT_VER_H_EXPLICIT_CAST_SPECIALISATIONS_REVISION   6
# define STLSOFT_VER_H_EXPLICIT_CAST_SPECIALISATIONS_EDIT       6
#endif /* !__STLSOFT_DOCUMENTATION_SKIP_SECTION */

/* /////////////////////////////////////////////////////////////////////////////
 * Auto-generation and compatibility
 */

/*
[<[STLSOFT-AUTO:NO-UNITTEST]>]
[<[STLSOFT-AUTO:NO-DOCFILELABEL]>]
*/

/* /////////////////////////////////////////////////////////////////////////////
 * Specialisations
 */

STLSOFT_TEMPLATE_SPECIALISATION
class explicit_cast<char const &>
{
public:
    typedef char                value_type;
    typedef explicit_cast<char> class_type;

// Construction
public:
    explicit_cast(char const &t)
        : m_t(t)
    {}

// Conversions
public:
    operator char const & () const
    {
        return m_t;
    }

// Members
private:
    char const &m_t;
};

#ifdef __STLSOFT_CF_NATIVE_WCHAR_T_SUPPORT

STLSOFT_TEMPLATE_SPECIALISATION
class explicit_cast<wchar_t const &>
{
public:
    typedef wchar_t                value_type;
    typedef explicit_cast<wchar_t> class_type;

// Construction
public:
    explicit_cast(wchar_t const &t)
        : m_t(t)
    {}

// Conversions
public:
    operator wchar_t const & () const
    {
        return m_t;
    }

// Members
private:
    wchar_t const &m_t;
};

#endif /* __STLSOFT_CF_NATIVE_WCHAR_T_SUPPORT */

STLSOFT_TEMPLATE_SPECIALISATION
class explicit_cast<unsigned char const &>
{
public:
    typedef unsigned char                value_type;
    typedef explicit_cast<unsigned char> class_type;

// Construction
public:
    explicit_cast(unsigned char const &t)
        : m_t(t)
    {}

// Conversions
public:
    operator unsigned char const & () const
    {
        return m_t;
    }

// Members
private:
    unsigned char const &m_t;
};

STLSOFT_TEMPLATE_SPECIALISATION
class explicit_cast<signed char const &>
{
public:
    typedef signed char                value_type;
    typedef explicit_cast<signed char> class_type;

// Construction
public:
    explicit_cast(signed char const &t)
        : m_t(t)
    {}

// Conversions
public:
    operator signed char const & () const
    {
        return m_t;
    }

// Members
private:
    signed char const &m_t;
};

STLSOFT_TEMPLATE_SPECIALISATION
class explicit_cast<signed short const &>
{
public:
    typedef signed short                value_type;
    typedef explicit_cast<signed short> class_type;

// Construction
public:
    explicit_cast(signed short const &t)
        : m_t(t)
    {}

// Conversions
public:
    operator signed short const & () const
    {
        return m_t;
    }

// Members
private:
    signed short const &m_t;
};

STLSOFT_TEMPLATE_SPECIALISATION
class explicit_cast<unsigned short const &>
{
public:
    typedef unsigned short                value_type;
    typedef explicit_cast<unsigned short> class_type;

// Construction
public:
    explicit_cast(unsigned short const &t)
        : m_t(t)
    {}

// Conversions
public:
    operator unsigned short const & () const
    {
        return m_t;
    }

// Members
private:
    unsigned short const &m_t;
};

STLSOFT_TEMPLATE_SPECIALISATION
class explicit_cast<signed int const &>
{
public:
    typedef signed int                value_type;
    typedef explicit_cast<signed int> class_type;

// Construction
public:
    explicit_cast(signed int const &t)
        : m_t(t)
    {}

// Conversions
public:
    operator signed int const & () const
    {
        return m_t;
    }

// Members
private:
    signed int const &m_t;
};

STLSOFT_TEMPLATE_SPECIALISATION
class explicit_cast<unsigned int const &>
{
public:
    typedef unsigned int                value_type;
    typedef explicit_cast<unsigned int> class_type;

// Construction
public:
    explicit_cast(unsigned int const &t)
        : m_t(t)
    {}

// Conversions
public:
    operator unsigned int const & () const
    {
        return m_t;
    }

// Members
private:
    unsigned int const &m_t;
};

STLSOFT_TEMPLATE_SPECIALISATION
class explicit_cast<signed long const &>
{
public:
    typedef signed long                value_type;
    typedef explicit_cast<signed long> class_type;

// Construction
public:
    explicit_cast(signed long const &t)
        : m_t(t)
    {}

// Conversions
public:
    operator signed long const & () const
    {
        return m_t;
    }

// Members
private:
    signed long const &m_t;
};

STLSOFT_TEMPLATE_SPECIALISATION
class explicit_cast<unsigned long const &>
{
public:
    typedef unsigned long                value_type;
    typedef explicit_cast<unsigned long> class_type;

// Construction
public:
    explicit_cast(unsigned long const &t)
        : m_t(t)
    {}

// Conversions
public:
    operator unsigned long const & () const
    {
        return m_t;
    }

// Members
private:
    unsigned long const &m_t;
};

#ifdef STLSOFT_CF_64BIT_INT_IS_long_long

STLSOFT_TEMPLATE_SPECIALISATION
class explicit_cast<signed long long const &>
{
public:
    typedef signed long long                value_type;
    typedef explicit_cast<signed long long> class_type;

// Construction
public:
    explicit_cast(signed long long const &t)
        : m_t(t)
    {}

// Conversions
public:
    operator signed long long const & () const
    {
        return m_t;
    }

// Members
private:
    signed long long const &m_t;
};

STLSOFT_TEMPLATE_SPECIALISATION
class explicit_cast<unsigned long long const &>
{
public:
    typedef unsigned long long                value_type;
    typedef explicit_cast<unsigned long long> class_type;

// Construction
public:
    explicit_cast(unsigned long long const &t)
        : m_t(t)
    {}

// Conversions
public:
    operator unsigned long long const & () const
    {
        return m_t;
    }

// Members
private:
    unsigned long long const &m_t;
};

#elif defined(STLSOFT_CF_64BIT_INT_IS___int64)

STLSOFT_TEMPLATE_SPECIALISATION
class explicit_cast<signed __int64 const &>
{
public:
    typedef signed __int64                value_type;
    typedef explicit_cast<signed __int64> class_type;

// Construction
public:
    explicit_cast(signed __int64 const &t)
        : m_t(t)
    {}

// Conversions
public:
    operator signed __int64 const & () const
    {
        return m_t;
    }

// Members
private:
    signed __int64 const &m_t;
};

STLSOFT_TEMPLATE_SPECIALISATION
class explicit_cast<unsigned __int64 const &>
{
public:
    typedef unsigned __int64                value_type;
    typedef explicit_cast<unsigned __int64> class_type;

// Construction
public:
    explicit_cast(unsigned __int64 const &t)
        : m_t(t)
    {}

// Conversions
public:
    operator unsigned __int64 const & () const
    {
        return m_t;
    }

// Members
private:
    unsigned __int64 const &m_t;
};

#else

# error 64-bit discrimination not handled correctly

#endif /* 64-bit */

STLSOFT_TEMPLATE_SPECIALISATION
class explicit_cast<float const &>
{
public:
    typedef float                value_type;
    typedef explicit_cast<float> class_type;

// Construction
public:
    explicit_cast(float const &t)
        : m_t(t)
    {}

// Conversions
public:
    operator float const & () const
    {
        return m_t;
    }

// Members
private:
    float const &m_t;
};

STLSOFT_TEMPLATE_SPECIALISATION
class explicit_cast<double const &>
{
public:
    typedef double                value_type;
    typedef explicit_cast<double> class_type;

// Construction
public:
    explicit_cast(double const &t)
        : m_t(t)
    {}

// Conversions
public:
    operator double const & () const
    {
        return m_t;
    }

// Members
private:
    double const &m_t;
};

STLSOFT_TEMPLATE_SPECIALISATION
class explicit_cast<long double const &>
{
public:
    typedef long double                value_type;
    typedef explicit_cast<long double> class_type;

// Construction
public:
    explicit_cast(long double const &t)
        : m_t(t)
    {}

// Conversions
public:
    operator long double const & () const
    {
        return m_t;
    }

// Members
private:
    long double const &m_t;
};

#ifdef __STLSOFT_CF_NATIVE_BOOL_SUPPORT

STLSOFT_TEMPLATE_SPECIALISATION
class explicit_cast<bool const &>
{
public:
    typedef bool                value_type;
    typedef explicit_cast<bool> class_type;

// Construction
public:
    explicit_cast(bool const &t)
        : m_t(t)
    {}

// Conversions
public:
    operator bool const & () const
    {
        return m_t;
    }

// Members
private:
    bool const &m_t;
};

#endif /* __STLSOFT_CF_NATIVE_BOOL_SUPPORT */

/* ////////////////////////////////////////////////////////////////////////// */

#endif /* !STLSOFT_INCL_H_STLSOFT_EXPLICIT_CAST */

/* ////////////////////////////////////////////////////////////////////////// */
