/* /////////////////////////////////////////////////////////////////////////////
 * File:        stlsoft_inheritance_disambiguator.h
 *
 * Purpose:     Class to disambiguate between parent classes of same type.
 *
 * Created:     1st September 2002
 * Updated:     30th May 2004
 *
 * Author:      Matthew Wilson, Synesis Software Pty Ltd.
 *
 * License:     (Licensed under the Synesis Software Standard Source License)
 *
 *              Copyright (C) 2002-2004, Synesis Software Pty Ltd.
 *
 *              All rights reserved.
 *
 *              www:        http://www.synesis.com.au/stlsoft
 *                          http://www.stlsoft.org/
 *
 *              email:      submissions@stlsoft.org  for submissions
 *                          admin@stlsoft.org        for other enquiries
 *
 *              Redistribution and use in source and binary forms, with or
 *              without modification, are permitted provided that the following
 *              conditions are met:
 *
 *              (i) Redistributions of source code must retain the above
 *              copyright notice and contact information, this list of
 *              conditions and the following disclaimer.
 *
 *              (ii) Any derived versions of this software (howsoever modified)
 *              remain the sole property of Synesis Software.
 *
 *              (iii) Any derived versions of this software (howsoever modified)
 *              remain subject to all these conditions.
 *
 *              (iv) Neither the name of Synesis Software nor the names of any
 *              subdivisions, employees or agents of Synesis Software, nor the
 *              names of any other contributors to this software may be used to
 *              endorse or promote products derived from this software without
 *              specific prior written permission.
 *
 *              This source code is provided by Synesis Software "as is" and any
 *              warranties, whether expressed or implied, including, but not
 *              limited to, the implied warranties of merchantability and
 *              fitness for a particular purpose are disclaimed. In no event
 *              shall the Synesis Software be liable for any direct, indirect,
 *              incidental, special, exemplary, or consequential damages
 *              (including, but not limited to, procurement of substitute goods
 *              or services; loss of use, data, or profits; or business
 *              interruption) however caused and on any theory of liability,
 *              whether in contract, strict liability, or tort (including
 *              negligence or otherwise) arising in any way out of the use of
 *              this software, even if advised of the possibility of such
 *              damage.
 *
 * ////////////////////////////////////////////////////////////////////////// */


/// \file stlsoft_inheritance_disambiguator.h
///
/// Class to disambiguate between parent classes of same type.

#ifndef STLSOFT_INCL_H_STLSOFT_INHERITANCE_DISAMBIGUATOR
#define STLSOFT_INCL_H_STLSOFT_INHERITANCE_DISAMBIGUATOR

#ifndef __STLSOFT_DOCUMENTATION_SKIP_SECTION
# define STLSOFT_VER_H_STLSOFT_INHERITANCE_DISAMBIGUATOR_MAJOR      1
# define STLSOFT_VER_H_STLSOFT_INHERITANCE_DISAMBIGUATOR_MINOR      2
# define STLSOFT_VER_H_STLSOFT_INHERITANCE_DISAMBIGUATOR_REVISION   2
# define STLSOFT_VER_H_STLSOFT_INHERITANCE_DISAMBIGUATOR_EDIT       12
#endif /* !__STLSOFT_DOCUMENTATION_SKIP_SECTION */

/* /////////////////////////////////////////////////////////////////////////////
 * Includes
 */

#ifndef STLSOFT_INCL_H_STLSOFT
# include "stlsoft.h"   // Include the STLSoft root header
#endif /* !STLSOFT_INCL_H_STLSOFT */

/* /////////////////////////////////////////////////////////////////////////////
 * Namespace
 */

#ifndef _STLSOFT_NO_NAMESPACE
namespace stlsoft
{
#endif /* _STLSOFT_NO_NAMESPACE */

/* ////////////////////////////////////////////////////////////////////////////
 * Classes
 */

// class inheritance_disambiguator

/// This class is used to disambiguate base classes, thereby faciliating multiple inheritance from the same type
///
/// \param T The type to disambiguate
/// \param N A unique integer index used to disambiguate between the classes
template <ss_typename_param_k T, int N>
class inheritance_disambiguator
    : public T
{
public:
    /// The disambiguated class type
    typedef T                               parent_class_type;
    enum { unique = N };
    /// The type of the current parameterisation
    typedef inheritance_disambiguator<T, N> class_type;

// Construction
public:
#ifdef __STLSOFT_CF_MEMBER_TEMPLATE_FUNCTION_SUPPORT
    // For compilers that support member templates, the following constructors
    // are provided.

    /// Default constructor
    inheritance_disambiguator()
    {}

    /// Single parameter constructor
    template <ss_typename_param_k N1>
    ss_explicit_k inheritance_disambiguator(N1 n1)
        : parent_class_type(n1)
    {}

    /// Two parameter constructor
    template <ss_typename_param_k N1, ss_typename_param_k N2>
    inheritance_disambiguator(N1 n1, N2 n2)
        : parent_class_type(n1, n2)
    {}

    /// Three parameter constructor
    template <ss_typename_param_k N1, ss_typename_param_k N2, ss_typename_param_k N3>
    inheritance_disambiguator(N1 n1, N2 n2, N3 n3)
        : parent_class_type(n1, n2, n3)
    {}

    /// Four parameter constructor
    template <ss_typename_param_k N1, ss_typename_param_k N2, ss_typename_param_k N3, ss_typename_param_k N4>
    inheritance_disambiguator(N1 n1, N2 n2, N3 n3, N4 n4)
        : parent_class_type(n1, n2, n3, n4)
    {}

    /// Five parameter constructor
    template <ss_typename_param_k N1, ss_typename_param_k N2, ss_typename_param_k N3, ss_typename_param_k N4, ss_typename_param_k N5>
    inheritance_disambiguator(N1 n1, N2 n2, N3 n3, N4 n4, N5 n5)
        : parent_class_type(n1, n2, n3, n4, n5)
    {}

    /// Six parameter constructor
    template <ss_typename_param_k N1, ss_typename_param_k N2, ss_typename_param_k N3, ss_typename_param_k N4, ss_typename_param_k N5, ss_typename_param_k N6>
    inheritance_disambiguator(N1 n1, N2 n2, N3 n3, N4 n4, N5 n5, N6 n6)
        : parent_class_type(n1, n2, n3, n4, n5, n6)
    {}

    /// Seven parameter constructor
    template <ss_typename_param_k N1, ss_typename_param_k N2, ss_typename_param_k N3, ss_typename_param_k N4, ss_typename_param_k N5, ss_typename_param_k N6, ss_typename_param_k N7>
    inheritance_disambiguator(N1 n1, N2 n2, N3 n3, N4 n4, N5 n5, N6 n6, N7 n7)
        : parent_class_type(n1, n2, n3, n4, n5, n6, n7)
    {}

    /// Eight parameter constructor
    template <ss_typename_param_k N1, ss_typename_param_k N2, ss_typename_param_k N3, ss_typename_param_k N4, ss_typename_param_k N5, ss_typename_param_k N6, ss_typename_param_k N7, ss_typename_param_k N8>
    inheritance_disambiguator(N1 n1, N2 n2, N3 n3, N4 n4, N5 n5, N6 n6, N7 n7, N8 n8)
        : parent_class_type(n1, n2, n3, n4, n5, n6, n7, n8)
    {}
#endif // __STLSOFT_CF_MEMBER_TEMPLATE_FUNCTION_SUPPORT
};

/* ////////////////////////////////////////////////////////////////////////// */

#ifndef _STLSOFT_NO_NAMESPACE
} // namespace stlsoft
#endif /* _STLSOFT_NO_NAMESPACE */

/* ////////////////////////////////////////////////////////////////////////// */

#endif /* !STLSOFT_INCL_H_STLSOFT_INHERITANCE_DISAMBIGUATOR */

/* ////////////////////////////////////////////////////////////////////////// */
