/* /////////////////////////////////////////////////////////////////////////////
 * File:        stlsoft_placement_aid.h
 *
 * Purpose:     Contains the auto_destructor and auto_array_destructor template
 *              classes.
 *
 * Created:     9th January 2002
 * Updated:     30th May 2004
 *
 * Author:      Matthew Wilson, Synesis Software Pty Ltd.
 *
 * License:     (Licensed under the Synesis Software Standard Source License)
 *
 *              Copyright (C) 2002-2004, Synesis Software Pty Ltd.
 *
 *              All rights reserved.
 *
 *              www:        http://www.synesis.com.au/stlsoft
 *                          http://www.stlsoft.org/
 *
 *              email:      submissions@stlsoft.org  for submissions
 *                          admin@stlsoft.org        for other enquiries
 *
 *              Redistribution and use in source and binary forms, with or
 *              without modification, are permitted provided that the following
 *              conditions are met:
 *
 *              (i) Redistributions of source code must retain the above
 *              copyright notice and contact information, this list of
 *              conditions and the following disclaimer.
 *
 *              (ii) Any derived versions of this software (howsoever modified)
 *              remain the sole property of Synesis Software.
 *
 *              (iii) Any derived versions of this software (howsoever modified)
 *              remain subject to all these conditions.
 *
 *              (iv) Neither the name of Synesis Software nor the names of any
 *              subdivisions, employees or agents of Synesis Software, nor the
 *              names of any other contributors to this software may be used to
 *              endorse or promote products derived from this software without
 *              specific prior written permission.
 *
 *              This source code is provided by Synesis Software "as is" and any
 *              warranties, whether expressed or implied, including, but not
 *              limited to, the implied warranties of merchantability and
 *              fitness for a particular purpose are disclaimed. In no event
 *              shall the Synesis Software be liable for any direct, indirect,
 *              incidental, special, exemplary, or consequential damages
 *              (including, but not limited to, procurement of substitute goods
 *              or services; loss of use, data, or profits; or business
 *              interruption) however caused and on any theory of liability,
 *              whether in contract, strict liability, or tort (including
 *              negligence or otherwise) arising in any way out of the use of
 *              this software, even if advised of the possibility of such
 *              damage.
 *
 * ////////////////////////////////////////////////////////////////////////// */


/// \file stlsoft_placement_aid.h
///
/// Contains the auto_destructor and auto_array_destructor template classes.

#ifndef STLSOFT_INCL_H_STLSOFT_PLACEMENT_AID
#define STLSOFT_INCL_H_STLSOFT_PLACEMENT_AID

#ifndef __STLSOFT_DOCUMENTATION_SKIP_SECTION
# define STLSOFT_VER_H_STLSOFT_PLACEMENT_AID_MAJOR      1
# define STLSOFT_VER_H_STLSOFT_PLACEMENT_AID_MINOR      2
# define STLSOFT_VER_H_STLSOFT_PLACEMENT_AID_REVISION   2
# define STLSOFT_VER_H_STLSOFT_PLACEMENT_AID_EDIT       13
#endif /* !__STLSOFT_DOCUMENTATION_SKIP_SECTION */

/* /////////////////////////////////////////////////////////////////////////////
 * Includes
 */

#ifndef STLSOFT_INCL_H_STLSOFT
# include "stlsoft.h"                // Include the STLSoft root header
#endif /* !STLSOFT_INCL_H_STLSOFT */
#ifndef STLSOFT_INCL_H_STLSOFT_CONSTRAINTS
# include "stlsoft_constraints.h"    // constraints
#endif /* !STLSOFT_INCL_H_STLSOFT_CONSTRAINTS */
#include <new>

/* /////////////////////////////////////////////////////////////////////////////
 * Namespace
 */

#ifndef _STLSOFT_NO_NAMESPACE
namespace stlsoft
{
#endif /* _STLSOFT_NO_NAMESPACE */

/* /////////////////////////////////////////////////////////////////////////////
 * Classes
 */

// class placement_aid

/// Scopes the lifetime of an in-place object
///
/// \param T The type of the in-place object
template <ss_typename_param_k T>
class placement_aid
{
public:
    /// The value type
    typedef T                   value_type;
    /// The type of the current parameterisation
    typedef placement_aid<T>    class_type;

// Construction
public:
    /// Create an instance of the \c value_type at the given location
    ///
    /// \param pv The location at which to in-place construct the object instance
    ss_explicit_k placement_aid(void *pv)
        : m_t(*new(pv) T())
    {}

    /// In-place destroy the object instance
    ~placement_aid()
    {
        m_t.~T();
    }

// Accessors
public:
    /// Implicit conversion operator to a reference to the in-place object instance
    operator T &()
    {
        return m_t;
    }
    /// Implicit conversion operator to a non-mutable (const) reference to the in-place object instance
    operator T const &() const
    {
        return m_t;
    }

    /// Address-of operator, providing pointer access to the in-place object instance
    T *operator &()
    {
        return &m_t;
    }
    /// Address-of operator, providing non-mutable (const) pointer access to the in-place object instance
    T const *operator &() const
    {
        return &m_t;
    }

// Members
private:
    T   &m_t;

// Not to be implemented
private:
    placement_aid(class_type const &);
    class_type const &operator =(class_type const &);
};

/* ////////////////////////////////////////////////////////////////////////// */

#ifndef _STLSOFT_NO_NAMESPACE
} // namespace stlsoft
#endif /* _STLSOFT_NO_NAMESPACE */

/* ////////////////////////////////////////////////////////////////////////// */

#endif /* !STLSOFT_INCL_H_STLSOFT_PLACEMENT_AID */

/* ////////////////////////////////////////////////////////////////////////// */
