/* /////////////////////////////////////////////////////////////////////////////
 * File:        stlsoft_remove_from_scope.h
 *
 * Purpose:     Contains the auto_destructor and auto_array_destructor template
 *              classes.
 *
 * Created:     10th June 2002
 * Updated:     30th May 2004
 *
 * Author:      Matthew Wilson, Synesis Software Pty Ltd.
 *
 * License:     (Licensed under the Synesis Software Standard Source License)
 *
 *              Copyright (C) 2002-2004, Synesis Software Pty Ltd.
 *
 *              All rights reserved.
 *
 *              www:        http://www.synesis.com.au/stlsoft
 *                          http://www.stlsoft.org/
 *
 *              email:      submissions@stlsoft.org  for submissions
 *                          admin@stlsoft.org        for other enquiries
 *
 *              Redistribution and use in source and binary forms, with or
 *              without modification, are permitted provided that the following
 *              conditions are met:
 *
 *              (i) Redistributions of source code must retain the above
 *              copyright notice and contact information, this list of
 *              conditions and the following disclaimer.
 *
 *              (ii) Any derived versions of this software (howsoever modified)
 *              remain the sole property of Synesis Software.
 *
 *              (iii) Any derived versions of this software (howsoever modified)
 *              remain subject to all these conditions.
 *
 *              (iv) Neither the name of Synesis Software nor the names of any
 *              subdivisions, employees or agents of Synesis Software, nor the
 *              names of any other contributors to this software may be used to
 *              endorse or promote products derived from this software without
 *              specific prior written permission.
 *
 *              This source code is provided by Synesis Software "as is" and any
 *              warranties, whether expressed or implied, including, but not
 *              limited to, the implied warranties of merchantability and
 *              fitness for a particular purpose are disclaimed. In no event
 *              shall the Synesis Software be liable for any direct, indirect,
 *              incidental, special, exemplary, or consequential damages
 *              (including, but not limited to, procurement of substitute goods
 *              or services; loss of use, data, or profits; or business
 *              interruption) however caused and on any theory of liability,
 *              whether in contract, strict liability, or tort (including
 *              negligence or otherwise) arising in any way out of the use of
 *              this software, even if advised of the possibility of such
 *              damage.
 *
 * ////////////////////////////////////////////////////////////////////////// */


/// \file stlsoft_remove_from_scope.h
///
/// Contains the auto_destructor and auto_array_destructor template classes.

#ifndef STLSOFT_INCL_H_STLSOFT_REMOVE_FROM_SCOPE
#define STLSOFT_INCL_H_STLSOFT_REMOVE_FROM_SCOPE

#ifndef __STLSOFT_DOCUMENTATION_SKIP_SECTION
# define STLSOFT_VER_H_STLSOFT_REMOVE_FROM_SCOPE_MAJOR      1
# define STLSOFT_VER_H_STLSOFT_REMOVE_FROM_SCOPE_MINOR      3
# define STLSOFT_VER_H_STLSOFT_REMOVE_FROM_SCOPE_REVISION   2
# define STLSOFT_VER_H_STLSOFT_REMOVE_FROM_SCOPE_EDIT       20
#endif /* !__STLSOFT_DOCUMENTATION_SKIP_SECTION */

/* /////////////////////////////////////////////////////////////////////////////
 * Includes
 */

#ifndef STLSOFT_INCL_H_STLSOFT
# include "stlsoft.h"                // Include the STLSoft root header
#endif /* !STLSOFT_INCL_H_STLSOFT */

/* /////////////////////////////////////////////////////////////////////////////
 * Namespace
 */

#ifndef _STLSOFT_NO_NAMESPACE
namespace stlsoft
{
#endif /* _STLSOFT_NO_NAMESPACE */

/* /////////////////////////////////////////////////////////////////////////////
 * Classes
 */

// class remove_from_scope

/// This class is almost completely unusable, and is used as a placeholder or as
/// a name remover
class remove_from_scope
{
// Construction
public:
    /// Default constructor
    remove_from_scope()
    {}
    /// Destructor
    ~remove_from_scope() stlsoft_throw_0()
    {}

// Not to be implemented
private:
    remove_from_scope(remove_from_scope const &rhs);
    remove_from_scope const &operator =(remove_from_scope const &rhs);

    // This function exists merely to prevent GCC from warning about
    // op-new returning null rather than throwing an exception
    static void *op_new_(size_t )
    {
        return 0;
    }

    void *operator new(size_t si)
    {
        return op_new_(si);
    }
    void operator delete(void *)
    {}

    int operator !() const;
#ifdef __STLSOFT_CF_NATIVE_BOOL_SUPPORT
    operator bool() const;
#endif /* __STLSOFT_CF_NATIVE_BOOL_SUPPORT */
    operator void *() const;
    void *operator &() const;
#if defined(__STLSOFT_COMPILER_IS_GCC) || \
    defined(__STLSOFT_COMPILER_IS_INTEL) || \
    defined(__STLSOFT_COMPILER_IS_WATCOM)
    struct _{};
    _ *operator ->() const;
#else
    struct {} *operator ->() const;
#endif /* __SYNSOFT_DBS_DEBUG */
#ifdef __STLSOFT_CF_MEMBER_TEMPLATE_FUNCTION_SUPPORT
    template <ss_typename_param_k T>
    operator T &();
    template <ss_typename_param_k T>
    operator T const &() const;
#endif // __STLSOFT_CF_MEMBER_TEMPLATE_FUNCTION_SUPPORT
};

/* ////////////////////////////////////////////////////////////////////////// */

#ifndef _STLSOFT_NO_NAMESPACE
} // namespace stlsoft
#endif /* _STLSOFT_NO_NAMESPACE */

/* ////////////////////////////////////////////////////////////////////////// */

#endif /* !STLSOFT_INCL_H_STLSOFT_REMOVE_FROM_SCOPE */

/* ////////////////////////////////////////////////////////////////////////// */
