/* /////////////////////////////////////////////////////////////////////////////
 * File:        winstl_char_conversions.h (originally MLStrCnv.h, ::SynesisStd)
 *
 * Purpose:     Type conversions for Windows.
 *
 * Created:
 * Updated:     8th June 2004
 *
 * Author:      Matthew Wilson, Synesis Software Pty Ltd.
 *
 * License:     (Licensed under the Synesis Software Standard Source License)
 *
 *              Copyright (C) 2002-2004, Synesis Software Pty Ltd.
 *
 *              All rights reserved.
 *
 *              www:        http://www.synesis.com.au/winstl
 *                          http://www.winstl.org/
 *
 *              email:      submissions@winstl.org  for submissions
 *                          admin@winstl.org        for other enquiries
 *
 *              Redistribution and use in source and binary forms, with or
 *              without modification, are permitted provided that the following
 *              conditions are met:
 *
 *              (i) Redistributions of source code must retain the above
 *              copyright notice and contact information, this list of
 *              conditions and the following disclaimer.
 *
 *              (ii) Any derived versions of this software (howsoever modified)
 *              remain the sole property of Synesis Software.
 *
 *              (iii) Any derived versions of this software (howsoever modified)
 *              remain subject to all these conditions.
 *
 *              (iv) Neither the name of Synesis Software nor the names of any
 *              subdivisions, employees or agents of Synesis Software, nor the
 *              names of any other contributors to this software may be used to
 *              endorse or promote products derived from this software without
 *              specific prior written permission.
 *
 *              This source code is provided by Synesis Software "as is" and any
 *              warranties, whether expressed or implied, including, but not
 *              limited to, the implied warranties of merchantability and
 *              fitness for a particular purpose are disclaimed. In no event
 *              shall the Synesis Software be liable for any direct, indirect,
 *              incidental, special, exemplary, or consequential damages
 *              (including, but not limited to, procurement of substitute goods
 *              or services; loss of use, data, or profits; or business
 *              interruption) however caused and on any theory of liability,
 *              whether in contract, strict liability, or tort (including
 *              negligence or otherwise) arising in any way out of the use of
 *              this software, even if advised of the possibility of such
 *              damage.
 *
 * ////////////////////////////////////////////////////////////////////////// */


/// \file winstl_char_conversions.h
///
/// Type conversions for Windows.

#ifndef WINSTL_INCL_H_WINSTL_CHAR_CONVERSIONS
#define WINSTL_INCL_H_WINSTL_CHAR_CONVERSIONS

#ifndef __STLSOFT_DOCUMENTATION_SKIP_SECTION
# define WINSTL_VER_H_WINSTL_CHAR_CONVERSIONS_MAJOR       1
# define WINSTL_VER_H_WINSTL_CHAR_CONVERSIONS_MINOR       6
# define WINSTL_VER_H_WINSTL_CHAR_CONVERSIONS_REVISION    3
# define WINSTL_VER_H_WINSTL_CHAR_CONVERSIONS_EDIT        32
#endif /* !__STLSOFT_DOCUMENTATION_SKIP_SECTION */

/* /////////////////////////////////////////////////////////////////////////////
 * Compatibility
 */

/*
[Incompatibilies-start]
__STLSOFT_COMPILER_IS_GCC: __GNUC__<3
__STLSOFT_COMPILER_IS_MSVC: _MSC_VER<1200
[Incompatibilies-end]
 */

/* /////////////////////////////////////////////////////////////////////////////
 * Includes
 */

#ifndef WINSTL_INCL_H_WINSTL
# include "winstl.h"                        // Include the WinSTL root header
#endif /* !WINSTL_INCL_H_WINSTL */

#if defined(__STLSOFT_COMPILER_IS_GCC) && \
    __GNUC__ < 3
# error winstl_char_conversions.h is not compatible with GNU C++ prior to 3.0
#endif /* compiler */
#if defined(__STLSOFT_COMPILER_IS_MSVC) && \
    _MSC_VER < 1200
# error winstl_char_conversions.h is not compatible with Visual C++ 5.0 or earlier
#endif /* _MSC_VER < 1200 */


#ifndef WINSTL_INCL_H_WINSTL_PROCESSHEAP_ALLOCATOR
# include "winstl_processheap_allocator.h"  // winstl::processheap_allocator
#endif /* !WINSTL_INCL_H_WINSTL_PROCESSHEAP_ALLOCATOR */
#ifndef STLSOFT_INCL_H_STLSOFT_STRING_ACCESS
# include "stlsoft_string_access.h"         // stlsoft::c_str_ptr
#endif /* !STLSOFT_INCL_H_STLSOFT_STRING_ACCESS */
#ifndef WINSTL_INCL_H_WINSTL_STRING_ACCESS
# include "winstl_string_access.h"          // winstl::c_str_ptr
#endif /* !WINSTL_INCL_H_WINSTL_STRING_ACCESS */
#ifndef STLSOFT_INCL_H_STLSOFT_AUTO_BUFFER
# include "stlsoft_auto_buffer.h"           // stlsoft::auto_buffer
#endif /* !STLSOFT_INCL_H_STLSOFT_AUTO_BUFFER */
#ifdef STLSOFT_UNITTEST
# include <wchar.h>
#endif /* STLSOFT_UNITTEST */

/* ////////////////////////////////////////////////////////////////////////// */

#ifndef _WINSTL_NO_NAMESPACE
# if defined(_STLSOFT_NO_NAMESPACE) || \
     defined(__STLSOFT_DOCUMENTATION_SKIP_SECTION)
/* There is no stlsoft namespace, so must define ::winstl */
namespace winstl
{
# else
/* Define stlsoft::winstl_project */

namespace stlsoft
{

namespace winstl_project
{

# endif /* _STLSOFT_NO_NAMESPACE */
#endif /* !_WINSTL_NO_NAMESPACE */

#if !defined(__STLSOFT_COMPILER_IS_MWERKS)
stlsoft_ns_using(c_str_ptr)
stlsoft_ns_using(c_str_len)
#endif /* compiler */

/* /////////////////////////////////////////////////////////////////////////////
 * Classes
 */

/// Converts an ANST (\c char-based) string to a Unicode (\c whar_t-based) string
template <ws_size_t CCH>
class ansi2Unicode
    : public auto_buffer<ws_char_w_t, processheap_allocator<ws_char_w_t>, CCH>
{
private:
    typedef auto_buffer<ws_char_w_t, processheap_allocator<ws_char_w_t>, CCH>   parent_class_type;
public:
    typedef ws_char_w_t                                                         char_type;
    typedef ss_typename_type_k parent_class_type::size_type                     size_type;
    typedef ss_typename_type_k parent_class_type::pointer                       pointer;

public:
#ifdef __STLSOFT_CF_MEMBER_TEMPLATE_CTOR_SUPPORT
    template <ss_typename_param_k S>
    ss_explicit_k ansi2Unicode(S const &s)
#else
    ss_explicit_k ansi2Unicode(ws_char_a_t const *s)
#endif // __STLSOFT_CF_MEMBER_TEMPLATE_CTOR_SUPPORT
        : parent_class_type(stlsoft_ns_qual(c_str_len)(s) + 1)
    {
        prepare_(s);
    }

#ifdef __STLSOFT_CF_MEMBER_TEMPLATE_CTOR_SUPPORT
    template <ss_typename_param_k S>
    ansi2Unicode(S const &s, size_type cch)
#else
    ansi2Unicode(ws_char_a_t const *s, size_type cch)
#endif // __STLSOFT_CF_MEMBER_TEMPLATE_CTOR_SUPPORT
        : parent_class_type(cch + 1)
    {
        prepare_(s);
    }

// Implementation
private:
#ifdef __STLSOFT_CF_MEMBER_TEMPLATE_CTOR_SUPPORT
    template <ss_typename_param_k S>
    void prepare_(S const &s)
#else
    void prepare_(ws_char_a_t const *s)
#endif // __STLSOFT_CF_MEMBER_TEMPLATE_CTOR_SUPPORT
    {
        // If the auto_buffer failed to allocate the required memory, and
        // we're not in an exception-environment, then size() will be zero
        if(0 == parent_class_type::size())
        {
            // Since we know that auto_buffer's parameterising size must
            // always be greater that 0, then
            parent_class_type::data()[0] = '\0';
        }
        else
        {
            size_type   size    =   parent_class_type::size();
            pointer     p       =   parent_class_type::data();

            // Note: cannot use -1 for length, since s might be a type that
            // could change length
            p[(0 == ::MultiByteToWideChar(0, 0, stlsoft_ns_qual(c_str_ptr)(s), static_cast<int>(size), p, static_cast<int>(size))) ? 0 : size - 1] = '\0';
        }
    }

/// \name Accessors
/// @{
public:
    char_type const *c_str() const
    {
        return parent_class_type::data();
    }
/// @}

// Not to be implemented
private:
    ansi2Unicode(ansi2Unicode const &);
    ansi2Unicode &operator =(ansi2Unicode const &);
};

/// Converts a Unicode (\c whar_t-based) string to an ANST (\c char-based) string
template <ws_size_t CCH>
class unicode2Ansi
    : public auto_buffer<ws_char_a_t, processheap_allocator<ws_char_a_t>, CCH>
{
private:
    typedef auto_buffer<ws_char_a_t, processheap_allocator<ws_char_a_t>, CCH>   parent_class_type;
public:
    typedef ws_char_a_t                                                         char_type;
    typedef ss_typename_type_k parent_class_type::size_type                     size_type;
    typedef ss_typename_type_k parent_class_type::pointer                       pointer;

public:
#ifdef __STLSOFT_CF_MEMBER_TEMPLATE_CTOR_SUPPORT
    template <ss_typename_param_k S>
    ss_explicit_k unicode2Ansi(S const &s)
#else
    ss_explicit_k unicode2Ansi(ws_char_w_t const *s)
#endif // __STLSOFT_CF_MEMBER_TEMPLATE_CTOR_SUPPORT
        : parent_class_type(stlsoft_ns_qual(c_str_len)(s) + 1)
    {
        prepare_(s);
    }

#ifdef __STLSOFT_CF_MEMBER_TEMPLATE_CTOR_SUPPORT
    template <ss_typename_param_k S>
    ss_explicit_k unicode2Ansi(S const &s, size_type cch)
#else
    ss_explicit_k unicode2Ansi(ws_char_w_t const *s, size_type cch)
#endif // __STLSOFT_CF_MEMBER_TEMPLATE_CTOR_SUPPORT
        : parent_class_type(cch + 1)
    {
        prepare_(s);
    }

// Implementation
private:
#ifdef __STLSOFT_CF_MEMBER_TEMPLATE_CTOR_SUPPORT
    template <ss_typename_param_k S>
    void prepare_(S const &s)
#else
    void prepare_(ws_char_a_t const *s)
#endif // __STLSOFT_CF_MEMBER_TEMPLATE_CTOR_SUPPORT
    {
        // If the auto_buffer failed to allocate the required memory, and
        // we're not in an exception-environment, then size() will be zero
        if(0 == parent_class_type::size())
        {
            // Since we know that auto_buffer's parameterising size must
            // always be greater that 0, then
            parent_class_type::data()[0] = '\0';
        }
        else
        {
            size_type   size    =   parent_class_type::size();
            pointer     p       =   parent_class_type::data();

            // Note: cannot use -1 for length, since s might be a type that
            // could change length
            p[(0 == ::WideCharToMultiByte(0, 0, stlsoft_ns_qual(c_str_ptr)(s), static_cast<int>(size), p, static_cast<int>(size), NULL, NULL)) ? 0 : size - 1] = '\0';
        }
    }

/// \name Accessors
/// @{
public:
    char_type const *c_str() const
    {
        return parent_class_type::data();
    }
/// @}

// Not to be implemented
private:
    unicode2Ansi(unicode2Ansi const &);
    unicode2Ansi &operator =(unicode2Ansi const &);
};

/* ////////////////////////////////////////////////////////////////////////// */

typedef ansi2Unicode<256>     a2w;
typedef unicode2Ansi<256>     w2a;

/* /////////////////////////////////////////////////////////////////////////////
 * Shims
 */

template<   ws_size_t           CCH
        >
inline ws_char_w_t const *c_str_ptr_null(ansi2Unicode<CCH> const &b)
{
    return stlsoft_ns_qual(c_str_ptr_null)(b.c_str());
}

template<   ws_size_t           CCH
        >
inline ws_char_w_t const *c_str_ptr(ansi2Unicode<CCH> const &b)
{
    return stlsoft_ns_qual(c_str_ptr)(b.c_str());
}

template<   ws_size_t           CCH
        >
inline ws_size_t c_str_len(ansi2Unicode<CCH> const &b)
{
    return stlsoft_ns_qual(c_str_len)(b.c_str());
}

template<   ws_size_t           CCH
        >
inline ws_size_t c_str_size(ansi2Unicode<CCH> const &b)
{
    return stlsoft_ns_qual(c_str_size)(b.c_str());
}

template<   ss_typename_param_k S
        ,   ws_size_t           CCH
        >
inline S &operator <<(S & s, ansi2Unicode<CCH> const &b)
{
    s << b.c_str();

    return s;
}

template<   ws_size_t           CCH
        >
inline ws_char_a_t const *c_str_ptr_null(unicode2Ansi<CCH> const &b)
{
    return stlsoft_ns_qual(c_str_ptr_null)(b.c_str());
}

template<   ws_size_t           CCH
        >
inline ws_char_a_t const *c_str_ptr(unicode2Ansi<CCH> const &b)
{
    return stlsoft_ns_qual(c_str_ptr)(b.c_str());
}

template<   ws_size_t           CCH
        >
inline ws_size_t c_str_len(unicode2Ansi<CCH> const &b)
{
    return stlsoft_ns_qual(c_str_len)(b.c_str());
}

template<   ws_size_t           CCH
        >
inline ws_size_t c_str_size(unicode2Ansi<CCH> const &b)
{
    return stlsoft_ns_qual(c_str_size)(b.c_str());
}

template<   ss_typename_param_k S
        ,   ws_size_t           CCH
        >
inline S &operator <<(S & s, unicode2Ansi<CCH> const &b)
{
    s << b.c_str();

    return s;
}

////////////////////////////////////////////////////////////////////////////////
// Unit-testing

#ifdef STLSOFT_UNITTEST

namespace unittest
{
    ss_bool_t test_winstl_char_conversions(unittest_reporter *r)
    {
        using stlsoft::unittest::unittest_initialiser;

        ss_bool_t               bSuccess    =   true;

        unittest_initialiser    init(r, "WinSTL", "char_conversion", __FILE__);

        if(0 != strcmp("String #1", w2a(L"String #1")))
        {
            r->report("w2a failed ", __LINE__);
            bSuccess = false;
        }

        if(0 != wcscmp(L"String #1", a2w("String #1")))
        {
            r->report("a2w failed ", __LINE__);
            bSuccess = false;
        }

        return bSuccess;
    }

    unittest_registrar    unittest_winstl_char_conversions(test_winstl_char_conversions);

} // namespace unittest

#endif /* STLSOFT_UNITTEST */

/* ////////////////////////////////////////////////////////////////////////// */

#ifndef _WINSTL_NO_NAMESPACE
# if defined(_STLSOFT_NO_NAMESPACE) || \
     defined(__STLSOFT_DOCUMENTATION_SKIP_SECTION)
} // namespace winstl
# else
} // namespace winstl_project
} // namespace stlsoft
# endif /* _STLSOFT_NO_NAMESPACE */
#endif /* !_WINSTL_NO_NAMESPACE */

/* /////////////////////////////////////////////////////////////////////////////
 * Namespace
 *
 * The string access shims exist either in the stlsoft namespace, or in the
 * global namespace. This is required by the lookup rules.
 *
 */

#ifndef _WINSTL_NO_NAMESPACE
# if !defined(_STLSOFT_NO_NAMESPACE) && \
     !defined(__STLSOFT_DOCUMENTATION_SKIP_SECTION)
namespace stlsoft
{
# else /* ? _STLSOFT_NO_NAMESPACE */
/* There is no stlsoft namespace, so must define in the global namespace */
# endif /* !_STLSOFT_NO_NAMESPACE */

using ::winstl::c_str_ptr_null;

using ::winstl::c_str_ptr;

using ::winstl::c_str_len;

using ::winstl::c_str_size;

# if !defined(_STLSOFT_NO_NAMESPACE) && \
     !defined(__STLSOFT_DOCUMENTATION_SKIP_SECTION)
} // namespace stlsoft
# else /* ? _STLSOFT_NO_NAMESPACE */
/* There is no stlsoft namespace, so must define in the global namespace */
# endif /* !_STLSOFT_NO_NAMESPACE */
#endif /* !_WINSTL_NO_NAMESPACE */

/* ////////////////////////////////////////////////////////////////////////// */

#endif /* WINSTL_INCL_H_WINSTL_CHAR_CONVERSIONS */

/* ////////////////////////////////////////////////////////////////////////// */
