/* ////////////////////////////////////////////////////////////////////////////
 * File:        winstl_event.h
 *
 * Purpose:     evemt class, based on Windows EVENT.
 *
 * Date:        3rd July 2003
 * Updated:     30th May 2004
 *
 * Author:      Matthew Wilson, Synesis Software Pty Ltd.
 *
 * License:     (Licensed under the Synesis Software Standard Source License)
 *
 *              Copyright (C) 2002-2004, Synesis Software Pty Ltd.
 *
 *              All rights reserved.
 *
 *              www:        http://www.synesis.com.au/winstl
 *                          http://www.winstl.org/
 *
 *              email:      submissions@winstl.org  for submissions
 *                          admin@winstl.org        for other enquiries
 *
 *              Redistribution and use in source and binary forms, with or
 *              without modification, are permitted provided that the following
 *              conditions are met:
 *
 *              (i) Redistributions of source code must retain the above
 *              copyright notice and contact information, this list of
 *              conditions and the following disclaimer.
 *
 *              (ii) Any derived versions of this software (howsoever modified)
 *              remain the sole property of Synesis Software.
 *
 *              (iii) Any derived versions of this software (howsoever modified)
 *              remain subject to all these conditions.
 *
 *              (iv) Neither the name of Synesis Software nor the names of any
 *              subdivisions, employees or agents of Synesis Software, nor the
 *              names of any other contributors to this software may be used to
 *              endorse or promote products derived from this software without
 *              specific prior written permission.
 *
 *              This source code is provided by Synesis Software "as is" and any
 *              warranties, whether expressed or implied, including, but not
 *              limited to, the implied warranties of merchantability and
 *              fitness for a particular purpose are disclaimed. In no event
 *              shall the Synesis Software be liable for any direct, indirect,
 *              incidental, special, exemplary, or consequential damages
 *              (including, but not limited to, procurement of substitute goods
 *              or services; loss of use, data, or profits; or business
 *              interruption) however caused and on any theory of liability,
 *              whether in contract, strict liability, or tort (including
 *              negligence or otherwise) arising in any way out of the use of
 *              this software, even if advised of the possibility of such
 *              damage.
 *
 * ////////////////////////////////////////////////////////////////////////// */


/// \file winstl_event.h
///
/// evemt class, based on Windows EVENT.

#ifndef WINSTL_INCL_H_WINSTL_EVENT
#define WINSTL_INCL_H_WINSTL_EVENT

#ifndef __STLSOFT_DOCUMENTATION_SKIP_SECTION
# define WINSTL_VER_H_WINSTL_EVENT_MAJOR        1
# define WINSTL_VER_H_WINSTL_EVENT_MINOR        4
# define WINSTL_VER_H_WINSTL_EVENT_REVISION     3
# define WINSTL_VER_H_WINSTL_EVENT_EDIT         19
#endif /* !__STLSOFT_DOCUMENTATION_SKIP_SECTION */

/* /////////////////////////////////////////////////////////////////////////////
 * Includes
 */

#ifndef WINSTL_INCL_H_WINSTL
# include "winstl.h"    // Include the WinSTL root header
#endif /* !WINSTL_INCL_H_WINSTL */

/* /////////////////////////////////////////////////////////////////////////////
 * Namespace
 */

#ifndef _WINSTL_NO_NAMESPACE
# if defined(_STLSOFT_NO_NAMESPACE) || \
     defined(__STLSOFT_DOCUMENTATION_SKIP_SECTION)
/* There is no stlsoft namespace, so must define ::winstl */
namespace winstl
{
# else
/* Define stlsoft::winstl_project */

namespace stlsoft
{

namespace winstl_project
{

# endif /* _STLSOFT_NO_NAMESPACE */
#endif /* !_WINSTL_NO_NAMESPACE */

/* /////////////////////////////////////////////////////////////////////////////
 * Classes
 */

#ifdef __SYNSOFT_DBS_COMPILER_SUPPORTS_PRAGMA_MESSAGE
# pragma message(_sscomp_fileline_message("This needs to be parameterised with a winstl::system_resource_policy, which would control whether to throw if MX create fails"))
#endif /* __SYNSOFT_DBS_COMPILER_SUPPORTS_PRAGMA_MESSAGE */

/// Class which wraps the Win32 EVENT synchronisation object
class event
{
public:
    typedef event       class_type;

// Construction
public:
    /// Creates the event
    event(ws_bool_t bManualReset, ws_bool_t bInitialState) winstl_throw_0()
        : m_hevent(create_event_(NULL, bManualReset, bInitialState, static_cast<ws_char_a_t const *>(0)))
    {}
    /// Creates the event with the given name
    ss_explicit_k event(ws_char_a_t const *name, ws_bool_t bManualReset, ws_bool_t bInitialState) winstl_throw_0()
        : m_hevent(create_event_(NULL, bManualReset, bInitialState, name))
    {}
    /// Creates the event with the given name
    ss_explicit_k event(ws_char_w_t const *name, ws_bool_t bManualReset, ws_bool_t bInitialState) winstl_throw_0()
        : m_hevent(create_event_(NULL, bManualReset, bInitialState, name))
    {}
    /// Creates the event with the given name and security attributes
    ss_explicit_k event(ws_char_a_t const *name, ws_bool_t bManualReset, ws_bool_t bInitialState, LPSECURITY_ATTRIBUTES psa) winstl_throw_0()
        : m_hevent(create_event_(psa, bManualReset, bInitialState, name))
    {}
    /// Creates the event with the given name and security attributes
    ss_explicit_k event(ws_char_w_t const *name, ws_bool_t bManualReset, ws_bool_t bInitialState, LPSECURITY_ATTRIBUTES psa) winstl_throw_0()
        : m_hevent(create_event_(psa, bManualReset, bInitialState, name))
    {}
    /// Destroys the event instance
    ~event()
    {
        if(m_hevent != NULL)
        {
            ::CloseHandle(m_hevent);
        }
    }

// Accessors
public:
    HANDLE  get_handle() const
    {
        return m_hevent;
    }

// Implementation
private:
    static HANDLE create_event_(LPSECURITY_ATTRIBUTES psa, ws_bool_t bManualReset, ws_bool_t bInitialState, ws_char_a_t const *name)
    {
        return ::CreateEventA(psa, bManualReset, bInitialState, name);
    }
    static HANDLE create_event_(LPSECURITY_ATTRIBUTES psa, ws_bool_t bManualReset, ws_bool_t bInitialState, ws_char_w_t const *name)
    {
        return ::CreateEventW(psa, bManualReset, bInitialState, name);
    }

// Members
private:
    HANDLE  m_hevent;

// Not to be implemented
private:
    event(class_type const &rhs);
    event &operator =(class_type const &rhs);
};

/* /////////////////////////////////////////////////////////////////////////////
 * Access shims
 */

/// \weakgroup concepts STLSoft Concepts

/// \weakgroup concepts_shims Shims
/// \ingroup concepts

/// \weakgroup concepts_shims_sync_control Synchronisation Control Shims
/// \ingroup concepts_shims
/// \brief These \ref concepts_shims "shims" control the behaviour of synchronisation objects

/// \defgroup winstl_sync_attribute_shims Synchronisation Access Shims (WinSTL)
/// \ingroup WinSTL concepts_shims_sync_control
/// \brief These \ref concepts_shims "shims" access the characteristics of Win32 synchronisation objects
/// @{

inline HANDLE get_handle(event const &ev)
{
    return ev.get_handle();
}

inline HANDLE get_handle(HANDLE h)
{
    return h;
}

/// @} // end of group winstl_sync_attribute_shims

////////////////////////////////////////////////////////////////////////////////
// Unit-testing

#ifdef STLSOFT_UNITTEST

namespace unittest
{
    ss_bool_t test_winstl_event(unittest_reporter *r)
    {
        using stlsoft::unittest::unittest_initialiser;

        ss_bool_t               bSuccess    =   true;

        unittest_initialiser    init(r, "WinSTL", "event", __FILE__);

#if 0
        if(<<TODO>>)
        {
            r->report("<<TODO>> failed ", __LINE__);
            bSuccess = false;
        }
#endif /* 0 */

        return bSuccess;
    }

    unittest_registrar    unittest_winstl_event(test_winstl_event);

} // namespace unittest

#endif /* STLSOFT_UNITTEST */

/* ////////////////////////////////////////////////////////////////////////// */

#ifndef _WINSTL_NO_NAMESPACE
# if defined(_STLSOFT_NO_NAMESPACE) || \
     defined(__STLSOFT_DOCUMENTATION_SKIP_SECTION)
} // namespace winstl
# else
} // namespace winstl_project
} // namespace stlsoft
# endif /* _STLSOFT_NO_NAMESPACE */
#endif /* !_WINSTL_NO_NAMESPACE */

/* ////////////////////////////////////////////////////////////////////////// */

#endif /* !WINSTL_INCL_H_WINSTL_EVENT */

/* ////////////////////////////////////////////////////////////////////////// */
