/* /////////////////////////////////////////////////////////////////////////////
 * File:        winstl_scale_functions.h
 *
 * Purpose:     Contains scale functions.
 *
 * Created:     7th February 2002
 * Updated:     30th May 2004
 *
 * Author:      Matthew Wilson, Synesis Software Pty Ltd.
 *
 * License:     (Licensed under the Synesis Software Standard Source License)
 *
 *              Copyright (C) 2002-2004, Synesis Software Pty Ltd.
 *
 *              All rights reserved.
 *
 *              www:        http://www.synesis.com.au/winstl
 *                          http://www.winstl.org/
 *
 *              email:      submissions@winstl.org  for submissions
 *                          admin@winstl.org        for other enquiries
 *
 *              Redistribution and use in source and binary forms, with or
 *              without modification, are permitted provided that the following
 *              conditions are met:
 *
 *              (i) Redistributions of source code must retain the above
 *              copyright notice and contact information, this list of
 *              conditions and the following disclaimer.
 *
 *              (ii) Any derived versions of this software (howsoever modified)
 *              remain the sole property of Synesis Software.
 *
 *              (iii) Any derived versions of this software (howsoever modified)
 *              remain subject to all these conditions.
 *
 *              (iv) Neither the name of Synesis Software nor the names of any
 *              subdivisions, employees or agents of Synesis Software, nor the
 *              names of any other contributors to this software may be used to
 *              endorse or promote products derived from this software without
 *              specific prior written permission.
 *
 *              This source code is provided by Synesis Software "as is" and any
 *              warranties, whether expressed or implied, including, but not
 *              limited to, the implied warranties of merchantability and
 *              fitness for a particular purpose are disclaimed. In no event
 *              shall the Synesis Software be liable for any direct, indirect,
 *              incidental, special, exemplary, or consequential damages
 *              (including, but not limited to, procurement of substitute goods
 *              or services; loss of use, data, or profits; or business
 *              interruption) however caused and on any theory of liability,
 *              whether in contract, strict liability, or tort (including
 *              negligence or otherwise) arising in any way out of the use of
 *              this software, even if advised of the possibility of such
 *              damage.
 *
 * ////////////////////////////////////////////////////////////////////////// */


/// \file winstl_scale_functions.h
///
/// Contains scale functions.

#ifndef WINSTL_INCL_H_WINSTL_SCALE_FUNCTIONS
#define WINSTL_INCL_H_WINSTL_SCALE_FUNCTIONS

#ifndef __STLSOFT_DOCUMENTATION_SKIP_SECTION
# define WINSTL_VER_H_WINSTL_SCALE_FUNCTIONS_MAJOR      1
# define WINSTL_VER_H_WINSTL_SCALE_FUNCTIONS_MINOR      4
# define WINSTL_VER_H_WINSTL_SCALE_FUNCTIONS_REVISION   3
# define WINSTL_VER_H_WINSTL_SCALE_FUNCTIONS_EDIT       19
#endif /* !__STLSOFT_DOCUMENTATION_SKIP_SECTION */

/* /////////////////////////////////////////////////////////////////////////////
 * Includes
 */

#include <winstl.h>             // Include the WinSTL root header

/* /////////////////////////////////////////////////////////////////////////////
 * Namespace
 */

#ifndef _WINSTL_NO_NAMESPACE
# if defined(_STLSOFT_NO_NAMESPACE) || \
     defined(__STLSOFT_DOCUMENTATION_SKIP_SECTION)
/* There is no stlsoft namespace, so must define ::winstl */
namespace winstl
{
# else
/* Define stlsoft::winstl_project */

namespace stlsoft
{

namespace winstl_project
{

# endif /* _STLSOFT_NO_NAMESPACE */
#endif /* !_WINSTL_NO_NAMESPACE */

/* /////////////////////////////////////////////////////////////////////////////
 * calc_width()
 *
 * This attribute shim calculates the width of an entity.
 */

/// Calculates the width of a \c SIZE instance
inline ws_sint_t calc_width(SIZE const &size)
{
    return size.cx;
}

/// Calculates the width of a \c RECT instance
inline ws_sint_t calc_width(RECT const &rc)
{
    return rc.right - rc.left;
}

/* /////////////////////////////////////////////////////////////////////////////
 * calc_height()
 *
 * This attribute shim calculates the height of an entity.
 */

/// Calculates the height of a \c SIZE instance
inline ws_sint_t calc_height(SIZE const &size)
{
    return size.cy;
}

/// Calculates the height of a \c RECT instance
inline ws_sint_t calc_height(RECT const &rc)
{
    return rc.bottom - rc.top;
}

////////////////////////////////////////////////////////////////////////////////
// Unit-testing

#ifdef STLSOFT_UNITTEST

namespace unittest
{
    ss_bool_t test_winstl_scale_functions(unittest_reporter *r)
    {
        using stlsoft::unittest::unittest_initialiser;

        ss_bool_t               bSuccess    =   true;

        unittest_initialiser    init(r, "WinSTL", "scale_functions", __FILE__);

        SIZE    size    =   { 100, 100 };
        RECT    rect    =   { 0, 0, 100, 100 };

        if(calc_width(size) != 100)
        {
            r->report("calc_width(SIZE) failed ", __LINE__);
            bSuccess = false;
        }

        if(calc_width(rect) != 100)
        {
            r->report("calc_width(RECT) failed ", __LINE__);
            bSuccess = false;
        }

        if(calc_height(size) != 100)
        {
            r->report("calc_height(SIZE) failed ", __LINE__);
            bSuccess = false;
        }

        if(calc_height(rect) != 100)
        {
            r->report("calc_height(RECT) failed ", __LINE__);
            bSuccess = false;
        }

        return bSuccess;
    }

    unittest_registrar    unittest_winstl_scale_functions(test_winstl_scale_functions);

} // namespace unittest

#endif /* STLSOFT_UNITTEST */

/* ////////////////////////////////////////////////////////////////////////// */

#ifndef _WINSTL_NO_NAMESPACE
# if defined(_STLSOFT_NO_NAMESPACE) || \
     defined(__STLSOFT_DOCUMENTATION_SKIP_SECTION)
} // namespace winstl
# else
} // namespace winstl_project
} // namespace stlsoft
# endif /* _STLSOFT_NO_NAMESPACE */
#endif /* !_WINSTL_NO_NAMESPACE */

/* ////////////////////////////////////////////////////////////////////////// */

#endif /* WINSTL_INCL_H_WINSTL_SCALE_FUNCTIONS */

/* ////////////////////////////////////////////////////////////////////////// */
