/* /////////////////////////////////////////////////////////////////////////////
 * File:        winstl_window_functionals.h
 *
 * Purpose:     Window specific functionals.
 *
 * Created:     19th January 2001
 * Updated:     30th May 2004
 *
 * Author:      Matthew Wilson, Synesis Software Pty Ltd.
 *
 * License:     (Licensed under the Synesis Software Standard Source License)
 *
 *              Copyright (C) 2002-2004, Synesis Software Pty Ltd.
 *
 *              All rights reserved.
 *
 *              www:        http://www.synesis.com.au/winstl
 *                          http://www.winstl.org/
 *
 *              email:      submissions@winstl.org  for submissions
 *                          admin@winstl.org        for other enquiries
 *
 *              Redistribution and use in source and binary forms, with or
 *              without modification, are permitted provided that the following
 *              conditions are met:
 *
 *              (i) Redistributions of source code must retain the above
 *              copyright notice and contact information, this list of
 *              conditions and the following disclaimer.
 *
 *              (ii) Any derived versions of this software (howsoever modified)
 *              remain the sole property of Synesis Software.
 *
 *              (iii) Any derived versions of this software (howsoever modified)
 *              remain subject to all these conditions.
 *
 *              (iv) Neither the name of Synesis Software nor the names of any
 *              subdivisions, employees or agents of Synesis Software, nor the
 *              names of any other contributors to this software may be used to
 *              endorse or promote products derived from this software without
 *              specific prior written permission.
 *
 *              This source code is provided by Synesis Software "as is" and any
 *              warranties, whether expressed or implied, including, but not
 *              limited to, the implied warranties of merchantability and
 *              fitness for a particular purpose are disclaimed. In no event
 *              shall the Synesis Software be liable for any direct, indirect,
 *              incidental, special, exemplary, or consequential damages
 *              (including, but not limited to, procurement of substitute goods
 *              or services; loss of use, data, or profits; or business
 *              interruption) however caused and on any theory of liability,
 *              whether in contract, strict liability, or tort (including
 *              negligence or otherwise) arising in any way out of the use of
 *              this software, even if advised of the possibility of such
 *              damage.
 *
 * ////////////////////////////////////////////////////////////////////////// */


/// \file winstl_window_functionals.h
///
/// Window specific functionals.

#ifndef WINSTL_INCL_H_WINSTL_WINDOW_FUNCTIONALS
#define WINSTL_INCL_H_WINSTL_WINDOW_FUNCTIONALS

#ifndef __STLSOFT_DOCUMENTATION_SKIP_SECTION
# define WINSTL_VER_H_WINSTL_WINDOW_FUNCTIONALS_MAJOR       1
# define WINSTL_VER_H_WINSTL_WINDOW_FUNCTIONALS_MINOR       1
# define WINSTL_VER_H_WINSTL_WINDOW_FUNCTIONALS_REVISION    7
# define WINSTL_VER_H_WINSTL_WINDOW_FUNCTIONALS_EDIT        16
#endif /* !__STLSOFT_DOCUMENTATION_SKIP_SECTION */

/* /////////////////////////////////////////////////////////////////////////////
 * Includes
 */

#ifndef WINSTL_INCL_H_WINSTL
# include "winstl.h"                // Include the WinSTL root header
#endif /* !WINSTL_INCL_H_WINSTL */
#ifndef WINSTL_INCL_H_WINSTL_WINDOW_ACCESS
# include "winstl_window_access.h"  // get_hwnd()
#endif /* !WINSTL_INCL_H_WINSTL_WINDOW_ACCESS */
#ifndef _WINSTL_WINDOW_FUNCTIONALS_NO_STD
# include <functional>
#endif /* _WINSTL_WINDOW_FUNCTIONALS_NO_STD */

/* /////////////////////////////////////////////////////////////////////////////
 * Namespace
 */

#ifndef _WINSTL_NO_NAMESPACE
# if defined(_STLSOFT_NO_NAMESPACE) || \
     defined(__STLSOFT_DOCUMENTATION_SKIP_SECTION)
/* There is no stlsoft namespace, so must define ::winstl */
namespace winstl
{
# else
/* Define stlsoft::winstl_project */

namespace stlsoft
{

namespace winstl_project
{

# endif /* _STLSOFT_NO_NAMESPACE */
#endif /* !_WINSTL_NO_NAMESPACE */

/* ////////////////////////////////////////////////////////////////////////// */

/// \weakgroup libraries STLSoft Libraries
/// \brief The individual libraries

/// \weakgroup libraries_window Window Library
/// \ingroup libraries
/// \brief This library provides facilities for defining and manipulating GUI windows

/// \defgroup winstl_window_library Window Library (WinSTL)
/// \ingroup WinSTL libraries_window
/// \brief This library provides facilities for defining and manipulating Win32 GUI windows
/// @{

/* /////////////////////////////////////////////////////////////////////////////
 * Classes
 */

/// Predicate used to determine whether windows are visible
struct is_visible
#ifndef _WINSTL_WINDOW_FUNCTIONALS_NO_STD
  : public winstl_ns_qual_std(unary_function)<HWND, ws_bool_t>
#endif /* !_WINSTL_WINDOW_FUNCTIONALS_NO_STD */
{
public:
#ifndef _WINSTL_WINDOW_FUNCTIONALS_NO_STD
    ///
private:
    typedef winstl_ns_qual_std(unary_function)<HWND, ws_bool_t> parent_class_type;
public:
    /// The argument type
    typedef parent_class_type::argument_type                    argument_type;
    /// The result type
    typedef parent_class_type::result_type                      result_type;
#else
    /// The argument type
    typedef HWND                                                argument_type;
    /// The result type
    typedef ws_bool_t                                           result_type;
#endif /* _WINSTL_WINDOW_FUNCTIONALS_NO_STD */

// Operations
public:
    /// Function call operator which evaluates whether the given window is visible
    ws_bool_t operator ()(HWND hwnd)
    {
        return ::IsWindowVisible(hwnd) != false;
    }
#ifdef __STLSOFT_CF_MEMBER_TEMPLATE_FUNCTION_SUPPORT
    /// Function call operator which evaluates whether the given window is visible
    template <ss_typename_param_k W>
    ws_bool_t operator ()(W const &w)
    {
        return operator ()(get_hwnd(w));
    }
#endif // __STLSOFT_CF_MEMBER_TEMPLATE_FUNCTION_SUPPORT
};



/// Predicate used to determine whether windows are enabled
struct is_enabled
#ifndef _WINSTL_WINDOW_FUNCTIONALS_NO_STD
  : public winstl_ns_qual_std(unary_function)<HWND, ws_bool_t>
#endif /* !_WINSTL_WINDOW_FUNCTIONALS_NO_STD */
{
public:
#ifndef _WINSTL_WINDOW_FUNCTIONALS_NO_STD
private:
    typedef winstl_ns_qual_std(unary_function)<HWND, ws_bool_t> parent_class_type;
public:
    /// The argument type
    typedef parent_class_type::argument_type                    argument_type;
    /// The result type
    typedef parent_class_type::result_type                      result_type;
#else
    /// The argument type
    typedef HWND                                                argument_type;
    /// The result type
    typedef ws_bool_t                                           result_type;
#endif /* _WINSTL_WINDOW_FUNCTIONALS_NO_STD */

// Operations
public:
    /// Function call operator which evaluates whether the given window is enabled
    ws_bool_t operator ()(HWND hwnd)
    {
        return ::IsWindowEnabled(hwnd) != false;
    }
#ifdef __STLSOFT_CF_MEMBER_TEMPLATE_FUNCTION_SUPPORT
    /// Function call operator which evaluates whether the given window is enabled
    template <ss_typename_param_k W>
    ws_bool_t operator ()(W const &w)
    {
        return operator ()(get_hwnd(w));
    }
#endif // __STLSOFT_CF_MEMBER_TEMPLATE_FUNCTION_SUPPORT
};

/* ////////////////////////////////////////////////////////////////////////// */

/// @} // end of group winstl_window_library

/* ////////////////////////////////////////////////////////////////////////// */

#ifndef _WINSTL_NO_NAMESPACE
# if defined(_STLSOFT_NO_NAMESPACE) || \
     defined(__STLSOFT_DOCUMENTATION_SKIP_SECTION)
} // namespace winstl
# else
} // namespace winstl_project
} // namespace stlsoft
# endif /* _STLSOFT_NO_NAMESPACE */
#endif /* !_WINSTL_NO_NAMESPACE */

/* ////////////////////////////////////////////////////////////////////////// */

#endif /* WINSTL_INCL_H_WINSTL_WINDOW_FUNCTIONALS */

/* ////////////////////////////////////////////////////////////////////////// */
