/* /////////////////////////////////////////////////////////////////////////////
 * File:        winstl_windows_ident.h
 *
 * Purpose:     Windows identification.
 *
 * Created:     11th March 2004
 * Updated:     30th May 2004
 *
 * Author:      Matthew Wilson, Synesis Software Pty Ltd.
 *
 * License:     (Licensed under the Synesis Software Standard Source License)
 *
 *              Copyright (C) 2002-2004, Synesis Software Pty Ltd.
 *
 *              All rights reserved.
 *
 *              www:        http://www.synesis.com.au/winstl
 *                          http://www.winstl.org/
 *
 *              email:      submissions@winstl.org  for submissions
 *                          admin@winstl.org        for other enquiries
 *
 *              Redistribution and use in source and binary forms, with or
 *              without modification, are permitted provided that the following
 *              conditions are met:
 *
 *              (i) Redistributions of source code must retain the above
 *              copyright notice and contact information, this list of
 *              conditions and the following disclaimer.
 *
 *              (ii) Any derived versions of this software (howsoever modified)
 *              remain the sole property of Synesis Software.
 *
 *              (iii) Any derived versions of this software (howsoever modified)
 *              remain subject to all these conditions.
 *
 *              (iv) Neither the name of Synesis Software nor the names of any
 *              subdivisions, employees or agents of Synesis Software, nor the
 *              names of any other contributors to this software may be used to
 *              endorse or promote products derived from this software without
 *              specific prior written permission.
 *
 *              This source code is provided by Synesis Software "as is" and any
 *              warranties, whether expressed or implied, including, but not
 *              limited to, the implied warranties of merchantability and
 *              fitness for a particular purpose are disclaimed. In no event
 *              shall the Synesis Software be liable for any direct, indirect,
 *              incidental, special, exemplary, or consequential damages
 *              (including, but not limited to, procurement of substitute goods
 *              or services; loss of use, data, or profits; or business
 *              interruption) however caused and on any theory of liability,
 *              whether in contract, strict liability, or tort (including
 *              negligence or otherwise) arising in any way out of the use of
 *              this software, even if advised of the possibility of such
 *              damage.
 *
 * ////////////////////////////////////////////////////////////////////////// */


/// \file winstl_windows_ident.h
///
/// Windows identification.

#ifndef WINSTL_INCL_H_WINSTL_WINDOWS_IDENT
#define WINSTL_INCL_H_WINSTL_WINDOWS_IDENT

#ifndef __STLSOFT_DOCUMENTATION_SKIP_SECTION
# define WINSTL_VER_H_WINSTL_WINDOWS_IDENT_MAJOR        1
# define WINSTL_VER_H_WINSTL_WINDOWS_IDENT_MINOR        2
# define WINSTL_VER_H_WINSTL_WINDOWS_IDENT_REVISION     5
# define WINSTL_VER_H_WINSTL_WINDOWS_IDENT_EDIT         14
#endif /* !__STLSOFT_DOCUMENTATION_SKIP_SECTION */

/* /////////////////////////////////////////////////////////////////////////////
 * Includes
 */

#ifndef WINSTL_INCL_H_WINSTL
# include "winstl.h"                        // Include the WinSTL root header
#endif /* !WINSTL_INCL_H_WINSTL */
#ifndef WINSTL_INCL_H_WINSTL_FUNCTIONS
//# include "winstl_functions.h"            //
#endif /* !WINSTL_INCL_H_WINSTL_FUNCTIONS */
#ifndef STLSOFT_INCL_H_STLSOFT_AUTO_BUFFER
# include "stlsoft_auto_buffer.h"           // auto_buffer
#endif /* !STLSOFT_INCL_H_STLSOFT_AUTO_BUFFER */
#ifndef WINSTL_INCL_H_PROCESSHEAP_ALLOCATOR
# include "winstl_processheap_allocator.h"  //
#endif /* !WINSTL_INCL_H_PROCESSHEAP_ALLOCATOR */

/* /////////////////////////////////////////////////////////////////////////////
 * Namespace
 */

#ifndef _WINSTL_NO_NAMESPACE
# if defined(_STLSOFT_NO_NAMESPACE) || \
     defined(__STLSOFT_DOCUMENTATION_SKIP_SECTION)
/* There is no stlsoft namespace, so must define ::winstl */
namespace winstl
{
# else
/* Define stlsoft::winstl_project */

namespace stlsoft
{

namespace winstl_project
{

# endif /* _STLSOFT_NO_NAMESPACE */
#endif /* !_WINSTL_NO_NAMESPACE */

/* /////////////////////////////////////////////////////////////////////////////
 * Enumerations
 */

enum WindowIdent
{
        Generic
    ,   ListBox         =   11
    ,   ComboBox        =   12
    ,   ListView        =   13
};

/* /////////////////////////////////////////////////////////////////////////////
 * Enumerations
 */

inline WindowIdent GetWindowIdent(HWND hwnd)
{
    typedef auto_buffer<char, processheap_allocator<char>, 256> buffer_t;

    struct Ident
    {
        WindowIdent ident;
        char const  *name;
    };

    WindowIdent         ident   =   Generic;
    buffer_t            buffer(buffer_t::internal_size() / 2);
    static const Ident  s_idents[] =
    {
            {   ListBox,    "LISTBOX"       }
        ,   {   ComboBox,   "COMBOBOX"      }
        ,   {   ListView,   "SysListView32" }
    };

    for(; buffer.resize(2 * buffer.size()); )
    {
        ws_size_t cch = static_cast<ws_size_t>(::GetClassNameA(hwnd, buffer, static_cast<int>(buffer.size())));

        if(cch < buffer.size())
        {
            break;
        }
    }

    for(ws_size_t index = 0; index < sizeof(s_idents) / sizeof(s_idents[0]); ++index)
    {
        if(0 == lstrcmpiA(s_idents[index].name, buffer))
        {
            ident = s_idents[index].ident;
            break;
        }
    }

    return ident;
}


////////////////////////////////////////////////////////////////////////////////
// Unit-testing

#ifdef STLSOFT_UNITTEST

namespace unittest
{
    ss_bool_t test_winstl_windows_ident(unittest_reporter *r)
    {
        using stlsoft::unittest::unittest_initialiser;

        ss_bool_t               bSuccess    =   true;

        unittest_initialiser    init(r, "WinSTL", "windows_ident", __FILE__);

        HWND    hwndListBox     =   ::CreateWindowA("LISTBOX", "", 0, 0, 0, 0, 0, NULL, NULL, NULL, NULL);
        HWND    hwndComboBox    =   ::CreateWindowA("COMBOBOX", "", 0, 0, 0, 0, 0, NULL, NULL, NULL, NULL);
        HWND    hwndListView    =   ::CreateWindowA("SysListView32", "", 0, 0, 0, 0, 0, NULL, NULL, NULL, NULL);

        if( NULL != hwndListBox &&
            GetWindowIdent(hwndListBox) != ListBox)
        {
            r->report("Identification of listbox failed ", __LINE__);
            bSuccess = false;
        }

        if( NULL != hwndComboBox &&
            GetWindowIdent(hwndComboBox) != ComboBox)
        {
            r->report("Identification of combobox failed ", __LINE__);
            bSuccess = false;
        }

        if( NULL != hwndListView &&
            GetWindowIdent(hwndListView) != ListView)
        {
            r->report("Identification of listview failed ", __LINE__);
            bSuccess = false;
        }

        return bSuccess;
    }

    unittest_registrar    unittest_winstl_windows_ident(test_winstl_windows_ident);

} // namespace unittest

#endif /* STLSOFT_UNITTEST */

/* ////////////////////////////////////////////////////////////////////////// */

#ifndef _WINSTL_NO_NAMESPACE
# if defined(_STLSOFT_NO_NAMESPACE) || \
     defined(__STLSOFT_DOCUMENTATION_SKIP_SECTION)
} // namespace winstl
# else
} // namespace winstl_project
} // namespace stlsoft
# endif /* _STLSOFT_NO_NAMESPACE */
#endif /* !_WINSTL_NO_NAMESPACE */

/* ////////////////////////////////////////////////////////////////////////// */

#endif /* WINSTL_INCL_H_WINSTL_WINDOWS_IDENT */

/* ////////////////////////////////////////////////////////////////////////// */
